# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import sys
import FreeCAD
from FreeCAD import Vector

import ObjectsFem
import Part

from BOPTools import SplitFeatures
from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Magnetostatic - Elmer 2D",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["magnetization"],
        "solvers": ["elmer"],
        "material": "solid",
        "equations": ["magnetostatic"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.equation_magnetostatics_2D_elmer import setup
setup()

Magnetodynamic2D equation - Elmer solver

"""
    )


def setup(doc=None, solvertype="elmer"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects

    # wire defining the lower horse shoe end
    p1 = Vector(0.0, -200.0, 0.0)
    p2 = Vector(200.0, -200.0, 0.0)
    p3 = Vector(200.0, -100.0, 0.0)
    p4 = Vector(0.0, -100.0, 0.0)
    Horseshoe_lower = doc.addObject("Part::Feature", "Lower_End")
    Horseshoe_lower.Shape = Part.makeFace(Part.makePolygon([p1, p2, p3, p4, p1]))
    Horseshoe_lower.ViewObject.Visibility = False

    # wire defining the upper horse shoe end
    p1 = Vector(0.0, 100.0, 0.0)
    p2 = Vector(200.0, 100.0, 0.0)
    p3 = Vector(200.0, 200.0, 0.0)
    p4 = Vector(0.0, 200.0, 0.0)
    Horseshoe_upper = doc.addObject("Part::Feature", "Upper_End")
    Horseshoe_upper.Shape = Part.makeFace(Part.makePolygon([p1, p2, p3, p4, p1]))
    Horseshoe_upper.ViewObject.Visibility = False

    # the U-part of the horse shoe
    # credits: https://forum.freecad.org/viewtopic.php?p=663051#p663051
    vpairs = [
        [Vector(340.0, 200.0, 0.0), Vector(200.0, 200.0, 0.0)],
        [Vector(200.0, 200.0, 0.0), Vector(200.0, 100.0, 0.0)],
        [Vector(200.0, 100.0, 0.0), Vector(325.0, 100.0, 0.0)],
        [Vector(325.0, 100.0, 0.0), Vector(325.0, -100.0, 0.0)],
        [Vector(325.0, -100.0, 0.0), Vector(200.0, -100.0, 0.0)],
        [Vector(200.0, -100.0, 0.0), Vector(200.0, -200.0, 0.0)],
        [Vector(200.0, -200.0, 0.0), Vector(340.0, -200.0, 0.0)],
        [Vector(340.0, 200.0, 0.0), Vector(340.0, -200.0, 0.0)],
    ]
    typeId = [
        "Part::GeomLine",
        "Part::GeomLine",
        "Part::GeomLine",
        "Part::GeomBSplineCurve",
        "Part::GeomLine",
        "Part::GeomLine",
        "Part::GeomLine",
        "Part::GeomBSplineCurve",
    ]
    e3Poles = [
        Vector(325.0, 100.0, 0.0),
        Vector(400.0, 100.0, 0.0),
        Vector(400.0, 0.0, 0.0),
        Vector(400.0, -100.0, 0.0),
        Vector(325.0, -100.0, 0.0),
    ]
    e3Knots = [0.0, 0.5, 1.0]
    e3Mults = [4, 1, 4]
    e3Degree = 3
    e7Poles = [
        Vector(340.0, 200.0, 0.0),
        Vector(500.0, 200.0, 0.0),
        Vector(500.0, 0.0, 0.0),
        Vector(500.0, -200.0, 0.0),
        Vector(340.0, -200.0, 0.0),
    ]
    e7Knots = [0.0, 0.5, 1.0]
    e7Mults = [4, 1, 4]
    e7Degree = 3
    c3 = Part.BSplineCurve()
    c3.buildFromPolesMultsKnots(e3Poles, e3Mults, e3Knots, False, e3Degree)
    c7 = Part.BSplineCurve()
    c7.buildFromPolesMultsKnots(e7Poles, e7Mults, e7Knots, False, e7Degree)
    edges = [c3.toShape(), c7.toShape()]
    for i in range(len(typeId)):
        if typeId[i] == "Part::GeomLine":
            edges.append(Part.makeLine(*vpairs[i]))

    sedges = Part.__sortEdges__(edges)
    Horseshoe_U = doc.addObject("Part::Feature", "Horseshoe_U")
    Horseshoe_U.Shape = Part.Face(Part.Wire(sedges))

    # a circle defining later the air volume
    Air_Circle = doc.addObject("Part::Feature", "Air_Circle")
    Air_Circle.Shape = Part.Face(Part.Wire(Part.makeCircle(1500.0)))

    # a link of the Horseshoe_lower
    Horseshoe_lowerLink = doc.addObject("App::Link", "Link_Lower_End")
    Horseshoe_lowerLink.LinkTransform = True
    Horseshoe_lowerLink.LinkedObject = Horseshoe_lower
    Horseshoe_lowerLink.ViewObject.Visibility = False

    # a link of the Horseshoe_upper
    Horseshoe_upperLink = doc.addObject("App::Link", "Link_Upper_End")
    Horseshoe_upperLink.LinkTransform = True
    Horseshoe_upperLink.LinkedObject = Horseshoe_upper
    Horseshoe_upperLink.ViewObject.Visibility = False

    # a link of the Horseshoe_U
    Horseshoe_ULink = doc.addObject("App::Link", "Link_Horseshoe_U")
    Horseshoe_ULink.LinkTransform = True
    Horseshoe_ULink.LinkedObject = Horseshoe_U
    Horseshoe_ULink.ViewObject.Visibility = False

    # fusion of all links
    Fusion = doc.addObject("Part::MultiFuse", "Fusion")
    Fusion.Shapes = [Horseshoe_lowerLink, Horseshoe_upperLink, Horseshoe_ULink]
    Fusion.ViewObject.Visibility = False

    # cut all objects from Air wire to get volume of fluid
    Cut = doc.addObject("Part::Cut", "Cut_Air")
    Cut.Base = Air_Circle
    Cut.Tool = Fusion
    Cut.ViewObject.Visibility = False

    # BooleanFregments object to combine cut with rod
    BooleanFragments = SplitFeatures.makeBooleanFragments(name="BooleanFragments")
    BooleanFragments.Objects = [Horseshoe_lower, Horseshoe_upper, Horseshoe_U, Cut]

    # set view
    doc.recompute()
    if FreeCAD.GuiUp:
        BooleanFragments.ViewObject.Document.activeView().viewTop()
        BooleanFragments.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")
    if FreeCAD.GuiUp:
        import FemGui

        FemGui.setActiveAnalysis(analysis)

    # solver
    if solvertype == "elmer":
        solver_obj = ObjectsFem.makeSolverElmer(doc, "SolverElmer")
        solver_obj.CoordinateSystem = "Cartesian 2D"
        equation_magnetodynamic2D = ObjectsFem.makeEquationMagnetodynamic2D(doc, solver_obj)
        equation_magnetodynamic2D.CalculateMagneticFieldStrength = True
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
        return doc
    analysis.addObject(solver_obj)

    # materials

    # air around the horse shoe
    material_obj = ObjectsFem.makeMaterialFluid(doc, "Air")
    mat = material_obj.Material
    mat["Name"] = "Air"
    mat["Density"] = "1.204 kg/m^3"
    mat["ThermalConductivity"] = "0.02587 W/m/K"
    mat["ThermalExpansionCoefficient"] = "3.43e-3 1/K"
    mat["SpecificHeat"] = "1.01 kJ/kg/K"
    mat["ElectricalConductivity"] = "1e-12 S/m"
    mat["RelativePermeability"] = "1.0"
    mat["RelativePermittivity"] = "1.00059"
    material_obj.Material = mat
    material_obj.References = [(BooleanFragments, "Face4")]
    analysis.addObject(material_obj)

    # iron of the horse shoe
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Iron")
    mat = material_obj.Material
    mat["Name"] = "Iron Generic"
    mat["ElectricalConductivity"] = "10.3 MS/m"
    mat["RelativePermeability"] = "5000.0"
    material_obj.Material = mat
    material_obj.References = [
        (BooleanFragments, "Face1"),
        (BooleanFragments, "Face2"),
        (BooleanFragments, "Face3"),
    ]
    analysis.addObject(material_obj)

    # magnetization lower
    Magnetization_lower = ObjectsFem.makeConstraintMagnetization(doc, "Magnetization_Lower_End")
    Magnetization_lower.References = [(BooleanFragments, "Face1")]
    Magnetization_lower.Magnetization_re_1 = "-7500.0 A/m"
    Magnetization_lower.EnableMagnetization_1 = True
    analysis.addObject(Magnetization_lower)

    # magnetization upper
    Magnetization_upper = ObjectsFem.makeConstraintMagnetization(doc, "Magnetization_Upper_End")
    Magnetization_upper.References = [(BooleanFragments, "Face2")]
    Magnetization_upper.Magnetization_re_1 = "7500.0 A/m"
    Magnetization_upper.EnableMagnetization_1 = True
    analysis.addObject(Magnetization_upper)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = BooleanFragments
    femmesh_obj.CharacteristicLengthMax = "100.0 mm"
    femmesh_obj.ViewObject.Visibility = False

    # mesh_region
    mesh_region = ObjectsFem.makeMeshRegion(doc, femmesh_obj, name="MeshRegion")
    mesh_region.CharacteristicLength = "6.5 mm"
    mesh_region.References = [
        (BooleanFragments, "Face1"),
        (BooleanFragments, "Face2"),
        (BooleanFragments, "Face3"),
    ]
    mesh_region.ViewObject.Visibility = False

    # generate the mesh
    generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM TD View command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_TDView:
    def GetResources(self):
        return {
            "Pixmap": "BIM_InsertView",
            "MenuText": QT_TRANSLATE_NOOP("BIM_TDView", "New View"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_TDView",
                "Inserts a drawing view on a page.\n"
                "To choose where to insert the view when multiple pages are available,\n"
                "select both the view and the page before executing the command.",
            ),
            "Accel": "V, I",
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        import Draft

        sections = []
        page = None
        drafts = []
        for obj in FreeCADGui.Selection.getSelection():
            t = Draft.getType(obj)
            if t == "SectionPlane":
                sections.append(obj)
            elif t == "TechDraw::DrawPage":
                page = obj
            else:
                drafts.append(obj)
        if not page:
            pages = FreeCAD.ActiveDocument.findObjects(Type="TechDraw::DrawPage")
            if pages:
                page = pages[0]
        if (not page) or ((not sections) and (not drafts)):
            FreeCAD.Console.PrintError(
                translate(
                    "BIM",
                    "No section view, Draft object, or page found or selected in the document",
                )
                + "\n"
            )
            return
        FreeCAD.ActiveDocument.openTransaction("Create view")
        for section in sections:
            view = FreeCAD.ActiveDocument.addObject("TechDraw::DrawViewArch", "BIM view")
            view.Label = section.Label
            view.Source = section
            page.addView(view)
            if page.Scale:
                view.Scale = page.Scale
        for draft in drafts:
            view = FreeCAD.ActiveDocument.addObject("TechDraw::DrawViewDraft", "DraftView")
            view.Label = draft.Label
            view.Source = draft
            page.addView(view)
            if page.Scale:
                view.Scale = page.Scale
            if "ShapeMode" in draft.PropertiesList:
                draft.ShapeMode = "Shape"
            for child in draft.OutListRecursive:
                if "ShapeMode" in child.PropertiesList:
                    child.ShapeMode = "Shape"
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


FreeCADGui.addCommand("BIM_TDView", BIM_TDView())



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2023 luvtofish                                          *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENSE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import Path
import numpy

__title__ = "Tapping Path Generator"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Generates the Tapping toolpath for a single spotshape"
__contributors__ = "luvtofish (Dan Henderson)"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


def generate(
    edge,
    dwelltime=0.0,
    repeat=1,
    retractheight=None,
    righthand=True,
    pitch=None,
    spindle_speed=None,
):
    """
    Generates Gcode for tapping a single hole.

    Takes as input an edge. It assumes the edge is trivial with just two vectors.
    The edge must be aligned with the Z axes (Vector(0,0,1)) or it is an error.

    The first vertex of the edge will be the startpoint
    The second vertex of the edge will be the endpoint.
    All other vertices are ignored.

    additionally, you can pass in a dwelltime,and repeat value.

    These will result in appropriate G74 and G84 codes.

    """
    startPoint = edge.Vertexes[0].Point
    endPoint = edge.Vertexes[1].Point

    Path.Log.debug(startPoint)
    Path.Log.debug(endPoint)

    Path.Log.debug(numpy.isclose(startPoint.sub(endPoint).x, 0, rtol=1e-05, atol=1e-06))
    Path.Log.debug(numpy.isclose(startPoint.sub(endPoint).y, 0, rtol=1e-05, atol=1e-06))
    Path.Log.debug(endPoint)

    if repeat < 1:
        raise ValueError("repeat must be 1 or greater")

    if not type(repeat) is int:
        raise ValueError("repeat value must be an integer")

    if not type(dwelltime) is float:
        raise ValueError("dwelltime must be a float")

    if retractheight is not None and not type(retractheight) is float:
        raise ValueError("retractheight must be a float")

    if not (
        numpy.isclose(startPoint.sub(endPoint).x, 0, rtol=1e-05, atol=1e-06)
        and (numpy.isclose(startPoint.sub(endPoint).y, 0, rtol=1e-05, atol=1e-06))
    ):
        raise ValueError("edge is not aligned with Z axis")

    if startPoint.z < endPoint.z:
        raise ValueError("start point is below end point")

    cmdParams = {}
    cmdParams["X"] = startPoint.x
    cmdParams["Y"] = startPoint.y
    cmdParams["Z"] = endPoint.z
    cmdParams["R"] = retractheight if retractheight is not None else startPoint.z
    cmdParams["S"] = spindle_speed if spindle_speed is not None else 1.0  # Sanity default
    cmdParams["F"] = float(pitch) if pitch is not None else 100.0  # Sanity default

    if repeat < 1:
        raise ValueError("repeat must be 1 or greater")

    if not type(repeat) is int:
        raise ValueError("repeat value must be an integer")

    if repeat > 1:
        cmdParams["L"] = repeat

    if dwelltime > 0.0:
        cmdParams["P"] = dwelltime

    # Check if tool is lefthand or righthand, set appropriate G-code
    if not (righthand):
        cmd = "G74"
    else:
        cmd = "G84"

    return [Path.Command(cmd, cmdParams)]



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Vector as vec

import Part
from Part import makeLine as ln

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "RC Wall 2D",
        "meshtype": "face",
        "meshelement": "Tria6",
        "constraints": ["fixed", "force", "displacement"],
        "solvers": ["ccxtools"],
        "material": "reinforced",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.rc_wall_2d import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=33106&start=80#p296469

example from Harry's epic topic: Concrete branch ready for testing

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    v1 = vec(0, -2000, 0)
    v2 = vec(500, -2000, 0)
    v3 = vec(500, 0, 0)
    v4 = vec(3500, 0, 0)
    v5 = vec(3500, -2000, 0)
    v6 = vec(4000, -2000, 0)
    v7 = vec(4000, 2000, 0)
    v8 = vec(0, 2000, 0)
    l1 = ln(v1, v2)
    l2 = ln(v2, v3)
    l3 = ln(v3, v4)
    l4 = ln(v4, v5)
    l5 = ln(v5, v6)
    l6 = ln(v6, v7)
    l7 = ln(v7, v8)
    l8 = ln(v8, v1)
    geom_obj = doc.addObject("Part::Feature", "FIB_Wall")
    geom_obj.Shape = Part.Face(Part.Wire([l1, l2, l3, l4, l5, l6, l7, l8]))
    doc.recompute()
    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.SplitInputWriter = False
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
    analysis.addObject(solver_obj)

    # shell thickness
    thickness_obj = ObjectsFem.makeElementGeometry2D(doc, 150.0, "ShellThickness")
    analysis.addObject(thickness_obj)

    # material
    matrixprop = {}
    matrixprop["Name"] = "Concrete-EN-C35_45"
    matrixprop["YoungsModulus"] = "32000 MPa"
    matrixprop["PoissonRatio"] = "0.17"
    matrixprop["CompressiveStrength"] = "15.75 MPa"
    # make some hint on the possible angle units in material system
    matrixprop["AngleOfFriction"] = "30 deg"
    reinfoprop = {}
    reinfoprop["Name"] = "Reinforcement-FIB-B500"
    reinfoprop["YieldStrength"] = "315 MPa"
    # not an official FreeCAD material property
    reinfoprop["ReinforcementRatio"] = "0.0"
    material_reinforced = ObjectsFem.makeMaterialReinforced(doc, "MaterialReinforced")
    material_reinforced.Material = matrixprop
    material_reinforced.Reinforcement = reinfoprop
    analysis.addObject(material_reinforced)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, "Edge1"), (geom_obj, "Edge5")]
    analysis.addObject(con_fixed)

    # constraint force
    con_force = ObjectsFem.makeConstraintForce(doc, "ConstraintForce")
    con_force.References = [(geom_obj, "Edge7")]
    con_force.Force = "1000000.0 N"
    con_force.Direction = (geom_obj, ["Edge8"])
    con_force.Reversed = False
    analysis.addObject(con_force)

    # constraint displacement
    con_disp = ObjectsFem.makeConstraintDisplacement(doc, "ConstraintDisplacementPrescribed")
    con_disp.References = [(geom_obj, "Face1")]
    con_disp.zFree = False
    con_disp.zDisplacement = 0
    analysis.addObject(con_disp)

    # mesh
    from .meshes.mesh_rc_wall_2d_tria6 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Arch Commands"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchCommands
#  \ingroup ARCH
#  \brief Utility functions for the Arch Workbench
#
#  This module provides general functions used by Arch tools
#  and utility commands

import FreeCAD
import ArchComponent
import Draft
import DraftVecUtils

from FreeCAD import Vector
from draftutils import params
from draftutils.groups import is_group

if FreeCAD.GuiUp:
    from PySide import QtGui, QtCore
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    # \endcond


# module functions ###############################################


def getStringList(objects):
    """getStringList(objects): returns a string defining a list
    of objects"""
    result = "["
    for o in objects:
        if len(result) > 1:
            result += ","
        result += "FreeCAD.ActiveDocument." + o.Name
    result += "]"
    return result


def getDefaultColor(objectType):
    """getDefaultColor(string): returns a color value for the given object
    type (Wall, Structure, Window, WindowGlass)"""
    alpha = 1.0
    if objectType == "Wall":
        c = params.get_param_arch("WallColor")
    elif objectType == "Structure":
        c = params.get_param_arch("StructureColor")
    elif objectType == "WindowGlass":
        c = params.get_param_arch("WindowGlassColor")
        alpha = 1.0 - params.get_param_arch("WindowTransparency") / 100.0
    elif objectType == "Rebar":
        c = params.get_param_arch("RebarColor")
    elif objectType == "Panel":
        c = params.get_param_arch("PanelColor")
    elif objectType == "Space":
        c = params.get_param_arch("defaultSpaceColor")
    elif objectType == "Helpers":
        c = params.get_param_arch("ColorHelpers")
    elif objectType == "Construction":
        c = params.get_param("constructioncolor")
        alpha = 0.2
    else:
        c = params.get_param_view("DefaultShapeColor")
    r, g, b, _ = Draft.get_rgba_tuple(c)
    return (r, g, b, alpha)


def _usedForAttachment(host, obj):
    if not getattr(obj, "AttachmentSupport", []):
        return False
    for sub in obj.AttachmentSupport:
        if sub[0] == host:
            return True
    return False


def addComponents(objectsList, host):
    """addComponents(objectsList,hostObject): adds the given object or the objects
    from the given list as components to the given host Object. Use this for
    example to add windows to a wall, or to add walls to a cell or floor."""
    if not isinstance(objectsList, list):
        objectsList = [objectsList]
    hostType = Draft.getType(host)
    if hostType in ["Floor", "Building", "Site", "Project", "BuildingPart"]:
        for o in objectsList:
            host.addObject(o)
    elif hostType in [
        "Wall",
        "CurtainWall",
        "Structure",
        "Precast",
        "Window",
        "Roof",
        "Stairs",
        "StructuralSystem",
        "Panel",
        "Component",
        "Pipe",
    ]:
        import DraftGeomUtils

        outList = host.OutListRecursive
        a = host.Additions
        x = getattr(host, "Axes", [])
        for o in objectsList:
            if hasattr(o, "Shape"):
                if Draft.getType(o) == "Window":
                    if hasattr(o, "Hosts"):
                        if not host in o.Hosts:
                            g = o.Hosts
                            g.append(host)
                            o.Hosts = g
                elif o in outList:
                    FreeCAD.Console.PrintWarning(
                        translate(
                            "Arch", "Cannot add {0} as it is already referenced by {1}."
                        ).format(o.Label, host.Label)
                        + "\n"
                    )
                elif Draft.getType(o) == "Axis":
                    if not o in x:
                        x.append(o)
                elif DraftGeomUtils.isValidPath(o.Shape) and (hostType in ["Structure", "Precast"]):
                    if _usedForAttachment(host, o):
                        o.AttachmentSupport = None
                        o.MapMode = "Deactivated"
                    host.Tool = o
                elif not o in a:
                    if hasattr(o, "Shape"):
                        a.append(o)
        host.Additions = a
        if hasattr(host, "Axes"):
            host.Axes = x
    elif hostType in ["SectionPlane"]:
        a = host.Objects
        for o in objectsList:
            if not o in a:
                a.append(o)
        host.Objects = a
    elif host.isDerivedFrom("App::DocumentObjectGroup"):
        for o in objectsList:
            host.addObject(o)


def removeComponents(objectsList, host=None):
    """removeComponents(objectsList,[hostObject]): removes the given component or
    the components from the given list from their parents. If a host object is
    specified, this function will try adding the components as holes to the host
    object instead."""
    if not isinstance(objectsList, list):
        objectsList = [objectsList]
    if host:
        if Draft.getType(host) in [
            "Wall",
            "CurtainWall",
            "Structure",
            "Precast",
            "Window",
            "Roof",
            "Stairs",
            "StructuralSystem",
            "Panel",
            "Component",
            "Pipe",
        ]:
            if getattr(host, "Tool", None) in objectsList:
                host.Tool = None
            if hasattr(host, "Axes"):
                a = host.Axes
                for o in objectsList[:]:
                    if o in a:
                        a.remove(o)
                        objectsList.remove(o)
                host.Axes = a
            s = host.Subtractions
            for o in objectsList:
                if Draft.getType(o) == "Window":
                    if hasattr(o, "Hosts"):
                        if not host in o.Hosts:
                            g = o.Hosts
                            g.append(host)
                            o.Hosts = g
                elif not o in s:
                    s.append(o)
                    if FreeCAD.GuiUp:
                        if Draft.getType(o) != "Roof":
                            setAsSubcomponent(o)
                    # Avoid cyclic dependency via Attachment Support:
                    if hasattr(o, "Base") and o.Base:
                        objList = [o, o.Base]
                    else:
                        objList = [o]
                    for i in objList:
                        if _usedForAttachment(host, i):
                            FreeCAD.Console.PrintMessage(
                                translate(
                                    "Arch",
                                    "{0} is mapped to {1}, removing the former's "
                                    + "Attachment Support to avoid cyclic dependency.",
                                ).format(o.Label, host.Label)
                                + "\n"
                            )
                            i.AttachmentSupport = None
                            i.MapMode = "Deactivated"
            host.Subtractions = s
        elif Draft.getType(host) in ["SectionPlane"]:
            a = host.Objects
            for o in objectsList:
                if o in a:
                    a.remove(o)
            host.Objects = a
    else:
        for o in objectsList:
            if o.InList:
                h = o.InList[0]
                tp = Draft.getType(h)
                if tp in ["Floor", "Building", "Site", "BuildingPart"]:
                    c = h.Group
                    if o in c:
                        c.remove(o)
                        h.Group = c
                        o.ViewObject.show()
                elif tp in ["Wall", "Structure", "Precast"]:
                    a = h.Additions
                    s = h.Subtractions
                    if o in a:
                        a.remove(o)
                        h.Additions = a
                        o.ViewObject.show()
                    elif o in s:
                        s.remove(o)
                        h.Subtractions = s
                        o.ViewObject.show()
                    elif o == h.Base:
                        h.Base = None
                        if FreeCAD.GuiUp:
                            o.ViewObject.show()
                elif tp in ["SectionPlane"]:
                    a = h.Objects
                    if o in a:
                        a.remove(o)
                        h.Objects = a
            if hasattr(o, "Hosts") and Draft.getType(o) == "Window":
                o.Hosts = []


def makeComponent(baseobj=None, name=None, delete=False):
    """makeComponent([baseobj],[name],[delete]): creates an undefined, non-parametric BIM
    component from the given base object"""
    if not FreeCAD.ActiveDocument:
        FreeCAD.Console.PrintError("No active document. Aborting\n")
        return
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "Component")
    obj.Label = name if name else translate("Arch", "Component")
    ArchComponent.Component(obj)
    if FreeCAD.GuiUp:
        ArchComponent.ViewProviderComponent(obj.ViewObject)
    if baseobj:
        import Part

        if hasattr(baseobj, "Shape"):
            obj.Shape = baseobj.Shape
            obj.Placement = baseobj.Placement
            if delete:
                FreeCAD.ActiveDocument.removeObject(baseobj.Name)
            else:
                obj.Base = baseobj
                if FreeCAD.GuiUp:
                    baseobj.ViewObject.hide()
        elif isinstance(baseobj, Part.Shape):
            obj.Shape = baseobj
    Draft.select(obj)
    return obj


def cloneComponent(obj):
    """cloneComponent(obj): Creates a clone of an object as an undefined component"""
    c = makeComponent()
    c.CloneOf = obj
    c.Placement = obj.Placement
    c.Label = obj.Label
    if hasattr(obj, "Material"):
        if obj.Material:
            c.Material = obj.Material
    if hasattr(obj, "IfcAttributes"):
        if obj.IfcAttributes:
            c.IfcAttributes = obj.IfcAttributes
    Draft.select(c)
    return c


def setAsSubcomponent(obj):
    """Sets the given object properly to become a subcomponent (addition, subtraction)
    of an Arch component"""
    Draft.ungroup(obj)
    if params.get_param_arch("applyConstructionStyle"):
        if FreeCAD.GuiUp:
            color = getDefaultColor("Construction")
            if hasattr(obj.ViewObject, "LineColor"):
                obj.ViewObject.LineColor = color
            if hasattr(obj.ViewObject, "PointColor"):
                obj.ViewObject.PointColor = color
            if hasattr(obj.ViewObject, "ShapeColor"):
                obj.ViewObject.ShapeColor = color
            if hasattr(obj.ViewObject, "Transparency"):
                obj.ViewObject.Transparency = int(color[3] * 100)
            obj.ViewObject.hide()


def copyProperties(obj1, obj2):
    """copyProperties(obj1,obj2): Copies properties values from obj1 to obj2,
    when that property exists in both objects"""
    for prop in obj1.PropertiesList:
        if prop in obj2.PropertiesList:
            if not prop in ["Proxy", "Shape"]:
                setattr(obj2, prop, getattr(obj1, prop))
    if obj1.ViewObject and obj2.ViewObject:
        for prop in obj1.ViewObject.PropertiesList:
            if prop in obj2.ViewObject.PropertiesList:
                if not prop in ["Proxy", "Shape"]:
                    setattr(obj2.ViewObject, prop, getattr(obj1.ViewObject, prop))


def splitMesh(obj, mark=True):
    """splitMesh(object,[mark]): splits the given mesh object into separated components.
    If mark is False, nothing else is done. If True (default), non-manifold components
    will be painted in red."""
    if not obj.isDerivedFrom("Mesh::Feature"):
        return []
    basemesh = obj.Mesh
    comps = basemesh.getSeparateComponents()
    nlist = []
    if comps:
        basename = obj.Name
        FreeCAD.ActiveDocument.removeObject(basename)
        for c in comps:
            newobj = FreeCAD.ActiveDocument.addObject("Mesh::Feature", basename)
            newobj.Mesh = c
            if mark and (not (c.isSolid()) or c.hasNonManifolds()):
                newobj.ViewObject.ShapeColor = (1.0, 0.0, 0.0, 1.0)
            nlist.append(newobj)
        return nlist
    return [obj]


def makeFace(wires, method=2, cleanup=False):
    """makeFace(wires): makes a face from a list of wires, finding which ones are holes"""
    # print("makeFace: start:", wires)
    import Part
    import DraftGeomUtils

    if not isinstance(wires, list):
        if len(wires.Vertexes) < 3:
            raise
        return Part.Face(wires)
    elif len(wires) == 1:
        # import Draft;Draft.printShape(wires[0])
        if len(wires[0].Vertexes) < 3:
            raise
        return Part.Face(wires[0])

    wires = wires[:]

    # print("makeFace: inner wires found")
    ext = None
    max_length = 0
    # cleaning up rubbish in wires
    if cleanup:
        for i in range(len(wires)):
            wires[i] = DraftGeomUtils.removeInterVertices(wires[i])
        # print("makeFace: garbage removed")
    for w in wires:
        # we assume that the exterior boundary is that one with
        # the biggest bounding box
        if w.BoundBox.DiagonalLength > max_length:
            max_length = w.BoundBox.DiagonalLength
            ext = w
    # print("makeFace: exterior wire", ext)
    wires.remove(ext)

    if method == 1:
        # method 1: reverse inner wires
        # all interior wires mark a hole and must reverse
        # their orientation, otherwise Part.Face fails
        for w in wires:
            # print("makeFace: reversing", w)
            w.reverse()
            # make sure that the exterior wires comes as first in the list
        wires.insert(0, ext)
        # print("makeFace: done sorting", wires)
        if wires:
            return Part.Face(wires)
    else:
        # method 2: use the cut method
        mf = Part.Face(ext)
        # print("makeFace: external face:", mf)
        for w in wires:
            f = Part.Face(w)
            # print("makeFace: internal face:", f)
            mf = mf.cut(f)
        # print("makeFace: final face:", mf.Faces)
        return mf.Faces[0]


def closeHole(shape):
    """closeHole(shape): closes a hole in an open shape"""
    import DraftGeomUtils
    import Part

    # creating an edges lookup table
    lut = {}
    for face in shape.Faces:
        for edge in face.Edges:
            hc = edge.hashCode()
            if hc in lut:
                lut[hc] = lut[hc] + 1
            else:
                lut[hc] = 1
    # filter out the edges shared by more than one face
    bound = []
    for e in shape.Edges:
        if lut[e.hashCode()] == 1:
            bound.append(e)
    bound = Part.__sortEdges__(bound)
    try:
        nface = Part.Face(Part.Wire(bound))
        shell = Part.makeShell(shape.Faces + [nface])
        solid = Part.Solid(shell)
    except Part.OCCError:
        raise
    else:
        return solid


def getCutVolume(cutplane, shapes, clip=False, depth=None):
    """getCutVolume(cutplane,shapes,[clip,depth]): returns a cut face and a cut volume
    from the given shapes and the given cutting plane. If clip is True, the cutvolume will
    also cut off everything outside the cutplane projection. If depth is non-zero, geometry
    further than this distance will be clipped off"""
    if not shapes:
        return None, None, None
    if not cutplane.Faces:
        return None, None, None
    import Part

    if not isinstance(shapes, list):
        shapes = [shapes]
    # building boundbox
    bb = shapes[0].optimalBoundingBox()
    for sh in shapes[1:]:
        bb.add(sh.optimalBoundingBox())
    bb.enlarge(1)
    # building cutplane space
    um = vm = wm = 0
    try:
        if hasattr(cutplane, "Shape"):
            p = cutplane.Shape.copy().Faces[0]
        else:
            p = cutplane.copy().Faces[0]
    except Part.OCCError:
        FreeCAD.Console.PrintMessage(translate("Arch", "Invalid cut plane") + "\n")
        return None, None, None
    ce = p.CenterOfMass
    ax = p.normalAt(0, 0)
    prm_range = p.ParameterRange  # (uMin, uMax, vMin, vMax)
    u = p.valueAt(prm_range[0], 0).sub(p.valueAt(prm_range[1], 0)).normalize()
    v = u.cross(ax)
    if not bb.isCutPlane(ce, ax):
        # FreeCAD.Console.PrintMessage(translate("Arch","No objects are cut by the plane)+"\n")
        return None, None, None
    else:
        corners = [
            FreeCAD.Vector(bb.XMin, bb.YMin, bb.ZMin),
            FreeCAD.Vector(bb.XMin, bb.YMax, bb.ZMin),
            FreeCAD.Vector(bb.XMax, bb.YMin, bb.ZMin),
            FreeCAD.Vector(bb.XMax, bb.YMax, bb.ZMin),
            FreeCAD.Vector(bb.XMin, bb.YMin, bb.ZMax),
            FreeCAD.Vector(bb.XMin, bb.YMax, bb.ZMax),
            FreeCAD.Vector(bb.XMax, bb.YMin, bb.ZMax),
            FreeCAD.Vector(bb.XMax, bb.YMax, bb.ZMax),
        ]
        for c in corners:
            dv = c.sub(ce)
            um1 = DraftVecUtils.project(dv, u).Length
            um = max(um, um1)
            vm1 = DraftVecUtils.project(dv, v).Length
            vm = max(vm, vm1)
            wm1 = DraftVecUtils.project(dv, ax).Length
            wm = max(wm, wm1)
        vu = DraftVecUtils.scaleTo(u, um)
        vui = vu.negative()
        vv = DraftVecUtils.scaleTo(v, vm)
        vvi = vv.negative()
        p1 = ce.add(vu.add(vvi))
        p2 = ce.add(vu.add(vv))
        p3 = ce.add(vui.add(vv))
        p4 = ce.add(vui.add(vvi))
        cutface = Part.makePolygon([p1, p2, p3, p4, p1])
        cutface = Part.Face(cutface)
        cutnormal = DraftVecUtils.scaleTo(ax, wm)
        cutvolume = cutface.extrude(cutnormal)
        cutnormal = cutnormal.negative()
        invcutvolume = cutface.extrude(cutnormal)
        if clip:
            extrudedplane = p.extrude(cutnormal)
            bordervolume = invcutvolume.cut(extrudedplane)
            cutvolume = cutvolume.fuse(bordervolume)
            cutvolume = cutvolume.removeSplitter()
            invcutvolume = extrudedplane
            cutface = p
        if depth:
            depthnormal = DraftVecUtils.scaleTo(cutnormal, depth)
            depthvolume = cutface.extrude(depthnormal)
            depthclipvolume = invcutvolume.cut(depthvolume)
            cutvolume = cutvolume.fuse(depthclipvolume)
            cutvolume = cutvolume.removeSplitter()
        return cutface, cutvolume, invcutvolume


def getShapeFromMesh(mesh, fast=True, tolerance=0.001, flat=False, cut=True):
    import Part
    import MeshPart
    import DraftGeomUtils

    if mesh.isSolid() and (mesh.countComponents() == 1) and fast:
        # use the best method
        faces = []
        for f in mesh.Facets:
            p = f.Points + [f.Points[0]]
            pts = []
            for pp in p:
                pts.append(FreeCAD.Vector(pp[0], pp[1], pp[2]))
            try:
                f = Part.Face(Part.makePolygon(pts))
            except Exception:
                print("getShapeFromMesh: error building face from polygon")
                # pass
            else:
                faces.append(f)
        shell = Part.makeShell(faces)
        try:
            solid = Part.Solid(shell)
        except Part.OCCError:
            print("getShapeFromMesh: error creating solid")
        else:
            try:
                solid = solid.removeSplitter()
            except Part.OCCError:
                print("getShapeFromMesh: error removing splitter")
                # pass
            return solid

    # if not mesh.isSolid():
    #    print "getShapeFromMesh: non-solid mesh, using slow method"
    faces = []
    segments = mesh.getPlanarSegments(tolerance)
    # print(len(segments))
    for i in segments:
        if len(i) > 0:
            wires = MeshPart.wireFromSegment(mesh, i)
            if wires:
                if flat:
                    nwires = []
                    for w in wires:
                        nwires.append(DraftGeomUtils.flattenWire(w))
                    wires = nwires
                try:
                    faces.append(makeFace(wires, method=int(cut) + 1))
                except Exception:
                    return None
    try:
        se = Part.makeShell(faces)
        se = se.removeSplitter()
        if flat:
            return se
    except Part.OCCError:
        print("getShapeFromMesh: error removing splitter")
        try:
            cp = Part.makeCompound(faces)
        except Part.OCCError:
            print("getShapeFromMesh: error creating compound")
            return None
        else:
            return cp
    else:
        try:
            solid = Part.Solid(se)
        except Part.OCCError:
            print("getShapeFromMesh: error creating solid")
            return se
        else:
            if solid.isClosed():
                return solid
            else:
                return se


def projectToVector(shape, vector):
    """projectToVector(shape,vector): projects the given shape on the given
    vector"""
    projpoints = []
    minl = 10000000000
    maxl = -10000000000
    for v in shape.Vertexes:
        p = DraftVecUtils.project(v.Point, vector)
        projpoints.append(p)
        l = p.Length
        if p.getAngle(vector) > 1:
            l = -l
        if l > maxl:
            maxl = l
        if l < minl:
            minl = l
    return DraftVecUtils.scaleTo(vector, maxl - minl)


def meshToShape(obj, mark=True, fast=True, tol=0.001, flat=False, cut=True):
    """meshToShape(object,[mark,fast,tol,flat,cut]): turns a mesh into a shape, joining coplanar facets. If
    mark is True (default), non-solid objects will be marked in red. Fast uses a faster algorithm by
    building a shell from the facets then removing splitter, tol is the tolerance used when converting
    mesh segments to wires, flat will force the wires to be perfectly planar, to be sure they can be
    turned into faces, but this might leave gaps in the final shell. If cut is true, holes in faces are
    made by subtraction (default)"""

    name = obj.Name
    if "Mesh" in obj.PropertiesList:
        mesh = obj.Mesh
        # plac = obj.Placement
        solid = getShapeFromMesh(mesh, fast, tol, flat, cut)
        if solid:
            if solid.isClosed() and solid.isValid():
                FreeCAD.ActiveDocument.removeObject(name)
            newobj = FreeCAD.ActiveDocument.addObject("Part::Feature", name)
            newobj.Shape = solid
            # newobj.Placement = plac #the placement is already computed in the mesh
            if (not solid.isClosed()) or (not solid.isValid()):
                if mark:
                    newobj.ViewObject.ShapeColor = (1.0, 0.0, 0.0, 1.0)
            return newobj
    return None


def removeCurves(shape, dae=False, tolerance=5):
    """removeCurves(shape,dae,tolerance=5): replaces curved faces in a shape
    with faceted segments. If dae is True, DAE triangulation options are used"""
    import Mesh

    if dae:
        from importers import importDAE

        t = importDAE.triangulate(shape.cleaned())
    else:
        t = shape.cleaned().tessellate(tolerance)
    m = Mesh.Mesh(t)
    return getShapeFromMesh(m)


def removeShape(objs, mark=True):
    """removeShape(objs,mark=True): takes an arch object (wall or structure) built on a cubic shape, and removes
    the inner shape, keeping its length, width and height as parameters. If mark is True, objects that cannot
    be processed by this function will become red."""
    import DraftGeomUtils

    if not isinstance(objs, list):
        objs = [objs]
    for obj in objs:
        if DraftGeomUtils.isCubic(obj.Shape):
            dims = DraftGeomUtils.getCubicDimensions(obj.Shape)
            if dims:
                name = obj.Name
                tp = Draft.getType(obj)
                print(tp)
                if tp == "Structure":
                    FreeCAD.ActiveDocument.removeObject(name)
                    import ArchStructure

                    str = ArchStructure.makeStructure(
                        length=dims[1], width=dims[2], height=dims[3], name=name
                    )
                    str.Placement = dims[0]
                elif tp == "Wall":
                    FreeCAD.ActiveDocument.removeObject(name)
                    import Arch

                    length = dims[1]
                    width = dims[2]
                    v1 = Vector(length / 2, 0, 0)
                    v2 = v1.negative()
                    v1 = dims[0].multVec(v1)
                    v2 = dims[0].multVec(v2)
                    line = Draft.makeLine(v1, v2)
                    Arch.makeWall(line, width=width, height=dims[3], name=name)
        else:
            if mark:
                obj.ViewObject.ShapeColor = (1.0, 0.0, 0.0, 1.0)


def mergeCells(objectslist):
    """mergeCells(objectslist): merges the objects in the given list
    into one. All objects must be of the same type and based on the Cell
    object (cells, floors, buildings, or sites)."""
    if not objectslist:
        return None
    if not isinstance(objectslist, list):
        return None
    if len(objectslist) < 2:
        return None
    typ = Draft.getType(objectslist[0])
    if not (typ in ["Cell", "Floor", "Building", "Site"]):
        return None
    for o in objectslist:
        if Draft.getType(o) != typ:
            return None
    base = objectslist.pop(0)
    for o in objectslist:
        l = base.Components
        for c in o.Components:
            if not c in l:
                l.append(c)
        base.Components = l
        FreeCAD.ActiveDocument.removeObject(o.Name)
    FreeCAD.ActiveDocument.recompute()
    return base


def download(url, force=False):
    """download(url,force=False): downloads a file from the given URL and saves it in the
    macro path. Returns the path to the saved file. If force is True, the file will be
    downloaded again evn if it already exists."""
    from urllib.request import urlopen
    import os

    name = url.split("/")[-1]
    macropath = FreeCAD.getUserMacroDir(True)
    filepath = os.path.join(macropath, name)
    if os.path.exists(filepath) and not (force):
        return filepath
    try:
        FreeCAD.Console.PrintMessage("Downloading " + url + " …\n")
        response = urlopen(url)
        s = response.read()
        f = open(filepath, "wb")
        f.write(s)
        f.close()
    except Exception:
        return None
    else:
        return filepath


def check(objectslist, includehidden=False):
    """check(objectslist,includehidden=False): checks if the given objects contain only solids"""
    objs = Draft.get_group_contents(objectslist)
    if not includehidden:
        objs = Draft.removeHidden(objs)
    bad = []
    for o in objs:
        if not hasattr(o, "Shape"):
            bad.append([o, "is not a Part-based object"])
        else:
            s = o.Shape
            if (not s.isClosed()) and (not (Draft.getType(o) == "Axis")):
                bad.append([o, translate("Arch", "is not closed")])
            elif not s.isValid():
                bad.append([o, translate("Arch", "is not valid")])
            elif (not s.Solids) and (not (Draft.getType(o) == "Axis")):
                bad.append([o, translate("Arch", "does not contain any solid")])
            else:
                f = 0
                for sol in s.Solids:
                    f += len(sol.Faces)
                    if not sol.isClosed():
                        bad.append([o, translate("Arch", "contains a non-closed solid")])
                if len(s.Faces) != f:
                    bad.append(
                        [o, translate("Arch", "contains faces that are not part of any solid")]
                    )
    return bad


def getHost(obj, strict=True):
    """getHost(obj,[strict]): returns the host of the current object. If strict is true (default),
    the host can only be an object of a higher level than the given one, or in other words, if a wall
    is contained in another wall which is part of a floor, the floor is returned instead of the parent wall
    """
    import Draft

    t = Draft.getType(obj)
    for par in obj.InList:
        if par.isDerivedFrom("Part::Feature") or par.isDerivedFrom("App::DocumentObjectGroup"):
            if strict:
                if Draft.getType(par) != t:
                    return par
                else:
                    return getHost(par, strict)
            else:
                return par
    return None


def pruneIncluded(objectslist, strict=False, silent=False):
    """pruneIncluded(objectslist,[strict]): removes from a list of Arch objects, those that are subcomponents of
    another shape-based object, leaving only the top-level shapes. If strict is True, the object
    is removed only if the parent is also part of the selection."""
    import Draft

    newlist = []
    for obj in objectslist:
        toplevel = True
        if obj.isDerivedFrom("Part::Feature"):
            if Draft.getType(obj) not in ["Window", "Clone", "Pipe", "Rebar", "Roof"]:
                for parent in obj.InList:
                    if not parent.isDerivedFrom("Part::Feature"):
                        pass
                    elif parent.isDerivedFrom("Part::Part2DObject"):
                        # don't consider 2D objects based on arch elements
                        pass
                    elif Draft.getType(parent) in [
                        "BezCurve",
                        "BSpline",
                        "Clone",
                        "Facebinder",
                        "Wire",
                        "Project",
                        "Roof",
                        "Site",
                        "Space",
                        "Window",
                    ]:
                        pass
                    elif parent.isDerivedFrom("PartDesign::FeatureBase"):
                        # don't consider a PartDesign_Clone that references obj
                        pass
                    elif parent.isDerivedFrom("PartDesign::Body") and obj == parent.BaseFeature:
                        # don't consider a PartDesign_Body with a PartDesign_Clone that references obj
                        pass
                    elif parent.isDerivedFrom("PartDesign::SubShapeBinder") or (
                        hasattr(parent, "TypeId") and parent.TypeId == "PartDesign::ShapeBinder"
                    ):
                        # don't consider a PartDesign_SubShapeBinder or PartDesign_ShapeBinder referencing this object from another object
                        pass
                    elif hasattr(parent, "Host") and parent.Host == obj:
                        pass
                    elif hasattr(parent, "Hosts") and obj in parent.Hosts:
                        pass
                    elif hasattr(parent, "TypeId") and parent.TypeId == "Part::Mirroring":
                        pass
                    elif hasattr(parent, "CloneOf"):
                        if parent.CloneOf:
                            if parent.CloneOf.Name != obj.Name:
                                toplevel = False
                        else:
                            toplevel = False
                    else:
                        toplevel = False

                    if toplevel == False and strict:
                        if parent not in objectslist and parent not in newlist:
                            toplevel = True
        if toplevel:
            newlist.append(obj)
        elif not silent:
            FreeCAD.Console.PrintWarning("pruning " + obj.Label + "\n")
    return newlist


def getAllChildren(objectlist):
    "getAllChildren(objectlist): returns all the children of all the object sin the list"
    obs = []
    for o in objectlist:
        if not o in obs:
            obs.append(o)
        if o.OutList:
            l = getAllChildren(o.OutList)
            for c in l:
                if not c in obs:
                    obs.append(c)
    return obs


def get_architectural_contents(
    initial_objects,
    recursive=True,
    discover_hosted_elements=True,
    include_components_from_additions=False,
    include_initial_objects_in_result=True,
):
    """
    Retrieves a flat list of unique architectural objects that are considered "contents" of or are
    related to the given initial_objects.

    This includes:
    - Children from .Group properties (hierarchical traversal if recursive=True).
    - Architecturally hosted elements (e.g., windows in walls, rebars in structures)
      if discover_hosted_elements=True.
    - Optionally, components from .Additions properties.

    The traversal uses a queue and ensures objects are processed only once.

    Parameters:
    -----------
    initial_objects : App::DocumentObject or list of App::DocumentObject
        The starting object(s) from which to discover contents.
    recursive : bool, optional
        If True (default), recursively find contents of discovered group-like
        objects (those with a .Group property and identified by draftutils.groups.is_group()).
    discover_hosted_elements : bool, optional
        If True (default), try to find elements like windows, doors, rebars
        that are architecturally hosted by other elements encountered during traversal.
        This relies on Draft.getType() and common Arch properties like .Hosts or .Host.
    include_components_from_additions : bool, optional
        If False (default), objects found in .Additions lists are not added.
        Set to True if these components should be part of the discovered contents.
    include_initial_objects_in_result : bool, optional
        If True (default), the objects in initial_objects themselves will be
        included in the output list (if not already discovered as a child of another).
        If False, only their "descendant" contents are returned.

    Returns:
    --------
    list of App::DocumentObject
        A flat list of unique architectural document objects.
    """

    final_contents_list = []
    queue = []

    if not isinstance(initial_objects, list):
        initial_objects_list = [initial_objects]
    else:
        initial_objects_list = list(initial_objects)  # Make a copy

    queue.extend(initial_objects_list)

    # Set to keep track of object names already added to the queue or fully processed
    # This prevents duplicates in the queue and reprocessing.
    processed_or_queued_names = set()
    for (
        item
    ) in initial_objects_list:  # Pre-populate for initial items if they are to be added later
        processed_or_queued_names.add(item.Name)

    idx = 0  # Use an index for iterating the queue, as it can grow
    while idx < len(queue):
        obj = queue[idx]
        idx += 1

        # Add the current object to the final list if it's not already there.
        # The decision to include initial_objects is handled by how they are first added to queue
        # and this check.
        if obj not in final_contents_list:
            is_initial = obj in initial_objects_list
            if (is_initial and include_initial_objects_in_result) or not is_initial:
                final_contents_list.append(obj)

        children_to_add_to_queue_next = []

        # 1. Hierarchical children from .Group (if recursive)
        if recursive and is_group(obj) and hasattr(obj, "Group") and obj.Group:
            for child in obj.Group:
                if child.Name not in processed_or_queued_names:
                    children_to_add_to_queue_next.append(child)
                    processed_or_queued_names.add(child.Name)  # Mark as queued

        # 2. Architecturally-hosted elements (if discover_hosted_elements)
        if discover_hosted_elements:
            host_types = ["Wall", "Structure", "CurtainWall", "Precast", "Panel", "Roof"]
            if Draft.getType(obj) in host_types:
                # Hosted elements are typically in the host's InList
                for item_in_inlist in obj.InList:
                    element_to_check = item_in_inlist
                    if hasattr(item_in_inlist, "getLinkedObject"):  # Resolve App::Link
                        linked = item_in_inlist.getLinkedObject()
                        if linked:
                            element_to_check = linked

                    if element_to_check.Name in processed_or_queued_names:
                        continue

                    is_confirmed_hosted = False
                    element_type = Draft.getType(element_to_check)

                    if element_type == "Window":  # This covers Arch Windows and Arch Doors
                        if hasattr(element_to_check, "Hosts") and obj in element_to_check.Hosts:
                            is_confirmed_hosted = True
                    elif element_type == "Rebar":
                        if hasattr(element_to_check, "Host") and obj == element_to_check.Host:
                            is_confirmed_hosted = True

                    if is_confirmed_hosted:
                        children_to_add_to_queue_next.append(element_to_check)
                        processed_or_queued_names.add(element_to_check.Name)  # Mark as queued

        # 3. Components from .Additions list (e.g., walls added to a main wall)
        if include_components_from_additions and hasattr(obj, "Additions") and obj.Additions:
            for addition_comp in obj.Additions:
                actual_addition = addition_comp
                if hasattr(addition_comp, "getLinkedObject"):  # Resolve if Addition is an App::Link
                    linked_add = addition_comp.getLinkedObject()
                    if linked_add:
                        actual_addition = linked_add

                if actual_addition.Name not in processed_or_queued_names:
                    children_to_add_to_queue_next.append(actual_addition)
                    processed_or_queued_names.add(actual_addition.Name)  # Mark as queued

        if children_to_add_to_queue_next:
            # Add newly-discovered children to the end of the queue. This function uses an index
            # (idx) to iterate through the queue, and queue.extend() adds new items to the end. This
            # results in a breadth-first-like traversal for objects discovered at the same "depth"
            # from different parent branches. Children of the current 'obj' will be processed after
            # 'obj's current siblings in the queue.
            queue.extend(children_to_add_to_queue_next)

    return final_contents_list


def survey(callback=False):
    """survey(): starts survey mode, where you can click edges and faces to get their lengths or area.
    Clicking on no object (on an empty area) resets the count."""
    if not callback:
        if hasattr(FreeCAD, "SurveyObserver"):
            for label in FreeCAD.SurveyObserver.labels:
                FreeCAD.ActiveDocument.removeObject(label)
            FreeCADGui.Selection.removeObserver(FreeCAD.SurveyObserver)
            del FreeCAD.SurveyObserver
            FreeCADGui.Control.closeDialog()
            if hasattr(FreeCAD, "SurveyDialog"):
                del FreeCAD.SurveyDialog
        else:
            FreeCAD.SurveyObserver = _SurveyObserver(callback=survey)
            FreeCADGui.Selection.addObserver(FreeCAD.SurveyObserver)
            FreeCAD.SurveyDialog = SurveyTaskPanel()
            FreeCADGui.Control.showDialog(FreeCAD.SurveyDialog)
    else:
        sel = FreeCADGui.Selection.getSelectionEx()
        if hasattr(FreeCAD, "SurveyObserver"):
            if not sel:
                if FreeCAD.SurveyObserver.labels:
                    for label in FreeCAD.SurveyObserver.labels:
                        FreeCAD.ActiveDocument.removeObject(label)
                    tl = FreeCAD.SurveyObserver.totalLength
                    ta = FreeCAD.SurveyObserver.totalArea
                    FreeCAD.SurveyObserver.labels = []
                    FreeCAD.SurveyObserver.selection = []
                    FreeCAD.SurveyObserver.totalLength = 0
                    FreeCAD.SurveyObserver.totalArea = 0
                    FreeCAD.SurveyObserver.totalVolume = 0
                    if not FreeCAD.SurveyObserver.cancellable:
                        FreeCAD.Console.PrintMessage("\n---- Reset ----\n\n")
                        FreeCAD.SurveyObserver.cancellable = True
                        if hasattr(FreeCAD, "SurveyDialog"):
                            FreeCAD.SurveyDialog.newline(tl, ta)
                    else:
                        FreeCADGui.Selection.removeObserver(FreeCAD.SurveyObserver)
                        del FreeCAD.SurveyObserver
                        FreeCADGui.Control.closeDialog()
                        if hasattr(FreeCAD, "SurveyDialog"):
                            del FreeCAD.SurveyDialog
            else:
                FreeCAD.SurveyObserver.cancellable = False
                basesel = FreeCAD.SurveyObserver.selection
                newsels = []
                for o in sel:
                    found = False
                    for eo in basesel:
                        if o.ObjectName == eo.ObjectName:
                            if o.SubElementNames == eo.SubElementNames:
                                found = True
                    if not found:
                        newsels.append(o)
                if newsels:
                    for o in newsels:
                        if hasattr(o.Object, "Shape"):
                            n = o.Object.Label
                            showUnit = params.get_param_arch("surveyUnits")
                            t = ""
                            u = FreeCAD.Units.Quantity()
                            if not o.HasSubObjects:
                                # entire object
                                anno = FreeCAD.ActiveDocument.addObject(
                                    "App::AnnotationLabel", "surveyLabel"
                                )
                                if hasattr(o.Object.Shape, "CenterOfMass"):
                                    anno.BasePosition = o.Object.Shape.CenterOfMass
                                else:
                                    anno.BasePosition = o.Object.Shape.BoundBox.Center
                                FreeCAD.SurveyObserver.labels.append(anno.Name)
                                if o.Object.Shape.Solids:
                                    u = FreeCAD.Units.Quantity(
                                        o.Object.Shape.Volume, FreeCAD.Units.Volume
                                    )
                                    t = u.getUserPreferred()[0]
                                    t = t.replace("^3", "³")
                                    anno.LabelText = "v " + t
                                    FreeCAD.Console.PrintMessage(
                                        "Object: " + n + ", Element: Whole, Volume: " + t + "\n"
                                    )
                                    FreeCAD.SurveyObserver.totalVolume += u.Value
                                elif o.Object.Shape.Faces:
                                    u = FreeCAD.Units.Quantity(
                                        o.Object.Shape.Area, FreeCAD.Units.Area
                                    )
                                    t = u.getUserPreferred()[0]
                                    t = t.replace("^2", "²")
                                    anno.LabelText = "a " + t
                                    FreeCAD.Console.PrintMessage(
                                        "Object: " + n + ", Element: Whole, Area: " + t + "\n"
                                    )
                                    FreeCAD.SurveyObserver.totalArea += u.Value
                                    if hasattr(FreeCAD, "SurveyDialog"):
                                        FreeCAD.SurveyDialog.update(2, t)
                                else:
                                    u = FreeCAD.Units.Quantity(
                                        o.Object.Shape.Length, FreeCAD.Units.Length
                                    )
                                    t = u.getUserPreferred()[0]
                                    anno.LabelText = "l " + t
                                    FreeCAD.Console.PrintMessage(
                                        "Object: " + n + ", Element: Whole, Length: " + t + "\n"
                                    )
                                    FreeCAD.SurveyObserver.totalLength += u.Value
                                    if hasattr(FreeCAD, "SurveyDialog"):
                                        FreeCAD.SurveyDialog.update(1, t)
                                if FreeCAD.GuiUp and t:
                                    if showUnit:
                                        QtGui.QApplication.clipboard().setText(t)
                                    else:
                                        QtGui.QApplication.clipboard().setText(str(u.Value))
                            else:
                                # single element(s)
                                for el in o.SubElementNames:
                                    e = getattr(o.Object.Shape, el)
                                    anno = FreeCAD.ActiveDocument.addObject(
                                        "App::AnnotationLabel", "surveyLabel"
                                    )
                                    if "Vertex" in el:
                                        anno.BasePosition = e.Point
                                    else:
                                        if hasattr(e, "CenterOfMass"):
                                            anno.BasePosition = e.CenterOfMass
                                        else:
                                            anno.BasePosition = e.BoundBox.Center
                                    FreeCAD.SurveyObserver.labels.append(anno.Name)
                                    if "Face" in el:
                                        u = FreeCAD.Units.Quantity(e.Area, FreeCAD.Units.Area)
                                        t = u.getUserPreferred()[0]
                                        t = t.replace("^2", "²")
                                        anno.LabelText = "a " + t
                                        FreeCAD.Console.PrintMessage(
                                            "Object: "
                                            + n
                                            + ", Element: "
                                            + el
                                            + ", Area: "
                                            + t
                                            + "\n"
                                        )
                                        FreeCAD.SurveyObserver.totalArea += u.Value
                                        if hasattr(FreeCAD, "SurveyDialog"):
                                            FreeCAD.SurveyDialog.update(2, t)
                                    elif "Edge" in el:
                                        u = FreeCAD.Units.Quantity(e.Length, FreeCAD.Units.Length)
                                        t = u.getUserPreferred()[0]
                                        anno.LabelText = "l " + t
                                        FreeCAD.Console.PrintMessage(
                                            "Object: "
                                            + n
                                            + ", Element: "
                                            + el
                                            + ", Length: "
                                            + t
                                            + "\n"
                                        )
                                        FreeCAD.SurveyObserver.totalLength += u.Value
                                        if hasattr(FreeCAD, "SurveyDialog"):
                                            FreeCAD.SurveyDialog.update(1, t)
                                    elif "Vertex" in el:
                                        u = FreeCAD.Units.Quantity(e.Z, FreeCAD.Units.Length)
                                        t = u.getUserPreferred()[0]
                                        anno.LabelText = "z " + t
                                        FreeCAD.Console.PrintMessage(
                                            "Object: "
                                            + n
                                            + ", Element: "
                                            + el
                                            + ", Zcoord: "
                                            + t
                                            + "\n"
                                        )
                                    if FreeCAD.GuiUp and t:
                                        if showUnit:
                                            QtGui.QApplication.clipboard().setText(t)
                                        else:
                                            QtGui.QApplication.clipboard().setText(str(u.Value))

                    FreeCAD.SurveyObserver.selection.extend(newsels)
            if hasattr(FreeCAD, "SurveyObserver"):
                if (
                    FreeCAD.SurveyObserver.totalLength
                    or FreeCAD.SurveyObserver.totalArea
                    or FreeCAD.SurveyObserver.totalVolume
                ):
                    msg = " Total:"
                    if FreeCAD.SurveyObserver.totalLength:
                        u = FreeCAD.Units.Quantity(
                            FreeCAD.SurveyObserver.totalLength, FreeCAD.Units.Length
                        )
                        t = u.getUserPreferred()[0]
                        msg += " Length: " + t
                    if FreeCAD.SurveyObserver.totalArea:
                        u = FreeCAD.Units.Quantity(
                            FreeCAD.SurveyObserver.totalArea, FreeCAD.Units.Area
                        )
                        t = u.getUserPreferred()[0]
                        t = t.replace("^2", "²")
                        msg += " Area: " + t
                    if FreeCAD.SurveyObserver.totalVolume:
                        u = FreeCAD.Units.Quantity(
                            FreeCAD.SurveyObserver.totalVolume, FreeCAD.Units.Volume
                        )
                        t = u.getUserPreferred()[0]
                        t = t.replace("^3", "³")
                        msg += " Volume: " + t
                    FreeCAD.Console.PrintMessage(msg + "\n")


class _SurveyObserver:
    "an observer for the survey() function"

    def __init__(self, callback):
        self.callback = callback
        self.selection = []
        self.labels = []
        self.totalLength = 0
        self.totalArea = 0
        self.totalVolume = 0
        self.cancellable = False
        self.doubleclear = False

    def addSelection(self, document, object, element, position):
        self.doubleclear = False
        self.callback(True)

    def clearSelection(self, document):
        if not self.doubleclear:
            self.doubleclear = True
        else:
            self.callback(True)


class SurveyTaskPanel:
    "A task panel for the survey tool"

    def __init__(self):
        self.form = QtGui.QWidget()
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Arch_Survey.svg"))
        layout = QtGui.QVBoxLayout(self.form)
        llayout = QtGui.QHBoxLayout()
        self.descr = QtGui.QLineEdit()
        llayout.addWidget(self.descr)
        self.addButton = QtGui.QPushButton()
        llayout.addWidget(self.addButton)
        layout.addLayout(llayout)
        self.tree = QtGui.QTreeWidget()
        self.tree.setColumnCount(3)
        layout.addWidget(self.tree)
        blayout = QtGui.QHBoxLayout()
        self.clearButton = QtGui.QPushButton()
        blayout.addWidget(self.clearButton)
        self.copyLength = QtGui.QPushButton()
        blayout.addWidget(self.copyLength)
        self.copyArea = QtGui.QPushButton()
        blayout.addWidget(self.copyArea)
        layout.addLayout(blayout)
        self.export = QtGui.QPushButton()
        layout.addWidget(self.export)
        QtCore.QObject.connect(self.addButton, QtCore.SIGNAL("clicked()"), self.setText)
        QtCore.QObject.connect(self.clearButton, QtCore.SIGNAL("clicked()"), self.clear)
        QtCore.QObject.connect(self.copyLength, QtCore.SIGNAL("clicked()"), self.clipLength)
        QtCore.QObject.connect(self.copyArea, QtCore.SIGNAL("clicked()"), self.clipArea)
        QtCore.QObject.connect(self.export, QtCore.SIGNAL("clicked()"), self.exportCSV)
        QtCore.QObject.connect(
            self.tree, QtCore.SIGNAL("itemClicked(QTreeWidgetItem*,int)"), self.setDescr
        )
        self.retranslateUi(self)
        item = QtGui.QTreeWidgetItem(self.tree)
        self.tree.setCurrentItem(item)

    def retranslateUi(self, dlg):
        self.form.setWindowTitle(QtGui.QApplication.translate("Arch", "Survey", None))
        self.addButton.setText(QtGui.QApplication.translate("Arch", "Set Description", None))
        self.clearButton.setText(QtGui.QApplication.translate("Arch", "Clear", None))
        self.copyLength.setText(QtGui.QApplication.translate("Arch", "Copy Total Length", None))
        self.copyArea.setText(QtGui.QApplication.translate("Arch", "Copy Total Area", None))
        self.export.setText(QtGui.QApplication.translate("Arch", "Export CSV", None))
        self.tree.setHeaderLabels(
            [
                QtGui.QApplication.translate("Arch", "Description", None),
                QtGui.QApplication.translate("Arch", "Length", None),
                QtGui.QApplication.translate("Arch", "Area", None),
            ]
        )

    def isAllowedAlterSelection(self):
        return True

    def isAllowedAlterView(self):
        return True

    def getStandardButtons(self):
        return QtGui.QDialogButtonBox.Close

    def reject(self):
        if hasattr(FreeCAD, "SurveyObserver"):
            for label in FreeCAD.SurveyObserver.labels:
                FreeCAD.ActiveDocument.removeObject(label)
            FreeCADGui.Selection.removeObserver(FreeCAD.SurveyObserver)
            del FreeCAD.SurveyObserver
        return True

    def clear(self):
        FreeCADGui.Selection.clearSelection()

    def clipLength(self):
        if hasattr(FreeCAD, "SurveyObserver"):
            u = FreeCAD.Units.Quantity(FreeCAD.SurveyObserver.totalLength, FreeCAD.Units.Length)
            t = u.getUserPreferred()[0]
            if params.get_param_arch("surveyUnits"):
                QtGui.QApplication.clipboard().setText(t)
            else:
                QtGui.QApplication.clipboard().setText(str(u.Value / u.getUserPreferred()[1]))

    def clipArea(self):
        if hasattr(FreeCAD, "SurveyObserver"):
            u = FreeCAD.Units.Quantity(FreeCAD.SurveyObserver.totalArea, FreeCAD.Units.Area)
            t = u.getUserPreferred()[0]
            t = t.replace("^2", "²")
            if params.get_param_arch("surveyUnits"):
                QtGui.QApplication.clipboard().setText(t)
            else:
                QtGui.QApplication.clipboard().setText(str(u.Value / u.getUserPreferred()[1]))

    def newline(self, length=0, area=0):
        FreeCADGui.Selection.clearSelection()
        item = QtGui.QTreeWidgetItem(self.tree)
        if length or area:
            item.setText(0, QtGui.QApplication.translate("Arch", "Total", None))
            item.setToolTip(0, "total")
            f = QtGui.QFont()
            f.setBold(True)
            item.setFont(0, f)
            item.setFont(1, f)
            item.setFont(2, f)
        else:
            item.setText(0, self.descr.text())
            self.descr.setText("")
        self.tree.setCurrentItem(item)
        if length:
            u = FreeCAD.Units.Quantity(length, FreeCAD.Units.Length)
            t = u.getUserPreferred()[0]
            item.setText(1, t)
        if area:
            u = FreeCAD.Units.Quantity(area, FreeCAD.Units.Area)
            t = u.getUserPreferred()[0]
            t = t.replace("^2", "²")
            item.setText(2, t)
        if length or area:
            item = QtGui.QTreeWidgetItem(self.tree)
            self.tree.setCurrentItem(item)

    def update(self, column, txt):
        item = QtGui.QTreeWidgetItem(self.tree)
        self.tree.setCurrentItem(item)
        item.setText(column, txt)

    def setDescr(self, item, col):
        self.descr.setText(item.text(0))

    def setText(self):
        item = self.tree.currentItem()
        if item:
            item.setText(0, self.descr.text())
            self.descr.setText("")

    def exportCSV(self):
        import csv

        rows = self.tree.topLevelItemCount()
        if rows:
            filename = QtGui.QFileDialog.getSaveFileName(
                QtGui.QApplication.activeWindow(),
                translate("Arch", "Export CSV File"),
                None,
                "CSV file (*.csv)",
            )
            if filename:
                with open(filename[0].encode("utf8"), "w") as csvfile:
                    csvfile = csv.writer(csvfile, delimiter="\t")
                    suml = 0
                    for i in range(rows):
                        item = self.tree.topLevelItem(i)
                        row = []
                        row.append(item.text(0))
                        if item.text(1):
                            u = FreeCAD.Units.Quantity(item.text(1))
                            if item.toolTip(0) == "total":
                                row.append("=SUM(B" + str(suml + 1) + ":B" + str(i) + ")")
                            else:
                                row.append(u.Value / u.getUserPreferred()[1])
                            row.append(u.getUserPreferred()[2])
                        else:
                            row.extend(["", ""])
                        if item.text(2):
                            t = item.text(2).replace("²", "^2")
                            u = FreeCAD.Units.Quantity(t)
                            if item.toolTip(0) == "total":
                                row.append("=SUM(D" + str(suml + 1) + ":D" + str(i) + ")")
                            else:
                                row.append(u.Value / u.getUserPreferred()[1])
                            row.append(u.getUserPreferred()[2])
                        else:
                            row.extend(["", ""])
                        csvfile.writerow(row)
                        if item.toolTip(0) == "total":
                            suml = i + 1
                print("successfully exported ", filename[0])


def toggleIfcBrepFlag(obj):
    """toggleIfcBrepFlag(obj): toggles the IFC brep flag of the given object, forcing it
    to be exported as brep geometry or not."""
    if not hasattr(obj, "IfcData"):
        FreeCAD.Console.PrintMessage(
            translate("Arch", "Object does not have settable IFC attributes")
        )
    else:
        d = obj.IfcData
        if "FlagForceBrep" in d:
            if d["FlagForceBrep"] == "True":
                d["FlagForceBrep"] = "False"
                FreeCAD.Console.PrintMessage(
                    translate("Arch", "Disabling B-rep force flag of object")
                    + " "
                    + obj.Label
                    + "\n"
                )
            else:
                d["FlagForceBrep"] = "True"
                FreeCAD.Console.PrintMessage(
                    translate("Arch", "Enabling B-rep force flag of object")
                    + " "
                    + obj.Label
                    + "\n"
                )
        else:
            d["FlagForceBrep"] = "True"
            FreeCAD.Console.PrintMessage(
                translate("Arch", "Enabling B-rep force flag of object") + " " + obj.Label + "\n"
            )
        obj.IfcData = d


def makeCompoundFromSelected(objects=None):
    """makeCompoundFromSelected([objects]): Creates a new compound object from the given
    subobjects (faces, edges) or from the selection if objects is None"""
    import FreeCADGui
    import Part

    so = []
    if not objects:
        objects = FreeCADGui.Selection.getSelectionEx()
    if not isinstance(objects, list):
        objects = [objects]
    for o in objects:
        so.extend(o.SubObjects)
    if so:
        c = Part.makeCompound(so)
        Part.show(c)


def cleanArchSplitter(objects=None):
    """cleanArchSplitter([objects]): removes the splitters from the base shapes
    of the given Arch objects or selected Arch objects if objects is None"""
    import FreeCAD
    import FreeCADGui

    if not objects:
        objects = FreeCADGui.Selection.getSelection()
    if not isinstance(objects, list):
        objects = [objects]
    for obj in objects:
        if hasattr(obj, "Shape"):
            if hasattr(obj, "Base"):
                if obj.Base:
                    print("Attempting to clean splitters from ", obj.Label)
                    base = obj.Base.getLinkedObject()
                    if base.isDerivedFrom("Part::Feature"):
                        if not base.Shape.isNull():
                            base.Shape = base.Shape.removeSplitter()
    FreeCAD.ActiveDocument.recompute()


def rebuildArchShape(objects=None):
    """rebuildArchShape([objects]): takes the faces from the base shape of the given (or selected
    if objects is None) Arch objects, and tries to rebuild a valid solid from them."""
    import FreeCAD
    import Part

    if not objects and FreeCAD.GuiUp:
        objects = FreeCADGui.Selection.getSelection()
    if not isinstance(objects, list):
        objects = [objects]
    for obj in objects:
        success = False
        if hasattr(obj, "Shape"):
            if hasattr(obj, "Base"):
                if obj.Base:
                    try:
                        print("Attempting to rebuild ", obj.Label)
                        base = obj.Base.getLinkedObject()
                        if base.isDerivedFrom("Part::Feature"):
                            if not base.Shape.isNull():
                                faces = []
                                for f in base.Shape.Faces:
                                    f2 = Part.Face(f.Wires)
                                    # print("rebuilt face: isValid is ", f2.isValid())
                                    faces.append(f2)
                                if faces:
                                    shell = Part.Shell(faces)
                                    if shell:
                                        # print("rebuilt shell: isValid is ", shell.isValid())
                                        solid = Part.Solid(shell)
                                        if solid:
                                            if not solid.isValid():
                                                solid.sewShape()
                                                solid = Part.Solid(solid)
                                            # print("rebuilt solid: isValid is ",solid.isValid())
                                            if solid.isValid():
                                                base.Shape = solid
                                                success = True
                    except Exception:
                        pass
        if not success:
            print("Failed to rebuild a valid solid for object ", obj.Name)
    FreeCAD.ActiveDocument.recompute()


def getExtrusionData(shape, sortmethod="area"):
    """If a shape has been extruded, returns the base face, and extrusion vector.

    Determines if a shape appears to have been extruded from some base face, and
    extruded at the normal from that base face. IE: it looks like a cuboid.
    https://en.wikipedia.org/wiki/Cuboid#Rectangular_cuboid

    If this is the case, returns what appears to be the base face, and the vector
    used to make that extrusion.

    The base face is determined based on the sortmethod parameter, which can either
    be:

    "area" = Of the faces with the smallest area, the one with the lowest z coordinate.
    "z" = The face with the lowest z coordinate.
    a 3D vector = the face which center is closest to the given 3D point

    Parameters
    ----------
    shape: <Part.Shape>
        Shape to examine.
    sortmethod: {"area", "z"}
        Which sorting algorithm to use to determine the base face.

    Returns
    -------
    Extrusion data: list
        Two item list containing the base face, and the vector used to create the
        extrusion. In that order.
    Failure: None
        Returns None if the object does not appear to be an extrusion.
    """

    if shape.isNull():
        return None
    if not shape.Solids:
        return None
    if len(shape.Faces) < 3:
        return None
    # build faces list with normals
    faces = []
    import Part

    for f in shape.Faces:
        try:
            faces.append([f, f.normalAt(0, 0)])
        except Part.OCCError:
            return None
    # find opposite normals pairs
    pairs = []
    for i1, f1 in enumerate(faces):
        for i2, f2 in enumerate(faces):
            if f1[0].hashCode() != f2[0].hashCode():
                if round(f1[1].getAngle(f2[1]), 4) == 3.1416:
                    pairs.append([i1, i2])
    if not pairs:
        return None
    valids = []
    for pair in pairs:
        hc = [faces[pair[0]][0].hashCode(), faces[pair[1]][0].hashCode()]
        # check if other normals are all at 90 degrees
        ok = True
        for f in faces:
            if f[0].hashCode() not in hc:
                if round(f[1].getAngle(faces[pair[0]][1]), 4) != 1.5708:
                    ok = False
        if ok:
            # prefer the face with the lowest z
            if faces[pair[0]][0].CenterOfMass.z < faces[pair[1]][0].CenterOfMass.z:
                valids.append(
                    [
                        faces[pair[0]][0],
                        faces[pair[1]][0].CenterOfMass.sub(faces[pair[0]][0].CenterOfMass),
                    ]
                )
            else:
                valids.append(
                    [
                        faces[pair[1]][0],
                        faces[pair[0]][0].CenterOfMass.sub(faces[pair[1]][0].CenterOfMass),
                    ]
                )
    if valids:
        if sortmethod == "z":
            valids.sort(key=lambda v: v[0].CenterOfMass.z)
        elif sortmethod == "area":
            # sort by smallest area
            valids.sort(key=lambda v: v[0].Area)
        else:
            valids.sort(key=lambda v: (v[0].CenterOfMass.sub(sortmethod)).Length)
        return valids[0]
    return None


def printMessage(message):
    FreeCAD.Console.PrintMessage(message)
    if FreeCAD.GuiUp:
        QtGui.QMessageBox.information(None, "", message)


def printWarning(message):
    FreeCAD.Console.PrintMessage(message)
    if FreeCAD.GuiUp:
        QtGui.QMessageBox.warning(None, "", message)


def makeIfcSpreadsheet(archobj=None):
    ifc_container = None
    for obj in FreeCAD.ActiveDocument.Objects:
        if obj.Name == "IfcPropertiesContainer":
            ifc_container = obj
    if not ifc_container:
        ifc_container = FreeCAD.ActiveDocument.addObject(
            "App::DocumentObjectGroup", "IfcPropertiesContainer"
        )
    import Spreadsheet

    ifc_spreadsheet = FreeCAD.ActiveDocument.addObject("Spreadsheet::Sheet", "IfcProperties")
    ifc_spreadsheet.set("A1", translate("Arch", "Category"))
    ifc_spreadsheet.set("B1", translate("Arch", "Key"))
    ifc_spreadsheet.set("C1", translate("Arch", "Type"))
    ifc_spreadsheet.set("D1", translate("Arch", "Value"))
    ifc_spreadsheet.set("E1", translate("Arch", "Unit"))
    ifc_container.addObject(ifc_spreadsheet)
    if archobj:
        if hasattr(obj, "IfcProperties"):
            archobj.IfcProperties = ifc_spreadsheet
            return ifc_spreadsheet
        else:
            FreeCAD.Console.PrintWarning(
                translate(
                    "Arch",
                    "The object does not have an IfcProperties attribute. Cancel spreadsheet creation for object:",
                )
                + " "
                + archobj.Label
            )
            FreeCAD.ActiveDocument.removeObject(ifc_spreadsheet)
    else:
        return ifc_spreadsheet



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Class to manage installation of sets of Python dependencies."""

import os
import subprocess
from typing import List

import addonmanager_freecad_interface as fci

# Get whatever version of PySide we can
try:
    from PySide import QtCore  # Use the FreeCAD wrapper
except ImportError:
    try:
        from PySide6 import QtCore  # Outside FreeCAD, try Qt6 first
    except ImportError:
        from PySide2 import QtCore  # Fall back to Qt5

import addonmanager_utilities as utils
from addonmanager_installer import AddonInstaller, MacroInstaller
from Addon import Addon

translate = fci.translate


class DependencyInstaller(QtCore.QObject):
    """Install Python dependencies using pip. Intended to be instantiated and then moved into a
    QThread: connect the run() function to the QThread's started() signal."""

    no_python_exe = QtCore.Signal()
    no_pip = QtCore.Signal(str)  # Attempted command
    failure = QtCore.Signal(str, str)  # Short message, detailed message
    finished = QtCore.Signal(bool)  # True if everything completed normally, otherwise false

    def __init__(
        self,
        addons: List[Addon],
        python_requires: List[str],
        python_optional: List[str],
        location: os.PathLike = None,
    ):
        """Install the various types of dependencies that might be specified. If an optional
         dependency fails this is non-fatal, but other failures are considered fatal. If location
        is specified it overrides the FreeCAD user base directory setting: this is used mostly
        for testing purposes and shouldn't be set by normal code in most circumstances.
        """
        super().__init__()
        self.addons = addons
        self.python_requires = python_requires
        self.python_optional = python_optional
        self.location = location
        self.required_succeeded = False
        self.finished_successfully = False

    def run(self):
        """Normally not called directly, but rather connected to the worker thread's started
        signal."""
        try:
            if self.python_requires or self.python_optional:
                if self._verify_pip():
                    if not QtCore.QThread.currentThread().isInterruptionRequested():
                        self._install_python_packages()
            else:
                self.required_succeeded = True
            if not QtCore.QThread.currentThread().isInterruptionRequested():
                self._install_addons()
                self.finished_successfully = self.required_succeeded
        except RuntimeError as e:
            fci.Console.PrintError(str(e) + "\n")
        self.finished.emit(self.finished_successfully)

    def _install_python_packages(self):
        """Install required and optional Python dependencies using pip."""

        if self.location:
            vendor_path = os.path.join(self.location, "AdditionalPythonPackages")
        else:
            vendor_path = utils.get_pip_target_directory()
        if not os.path.exists(vendor_path):
            os.makedirs(vendor_path)

        self.required_succeeded = self._install_required(vendor_path)
        self._install_optional(vendor_path)

    def _verify_pip(self) -> bool:
        """Ensure that pip is working -- returns True if it is, or False if not. Also emits the
        no_pip signal if pip cannot execute."""
        try:
            proc = self._run_pip(["--version"])
            fci.Console.PrintMessage(proc.stdout + "\n")
            if proc.returncode != 0:
                return False
        except subprocess.CalledProcessError:
            call = utils.create_pip_call([])
            self.no_pip.emit(" ".join(call))
            return False
        return True

    def _install_required(self, vendor_path: str) -> bool:
        """Install the required Python package dependencies. If any fail a failure
        signal is emitted and the function exits without proceeding with any additional
        installations."""
        for pymod in self.python_requires:
            if pymod.lower().startswith("pyside"):
                continue  # Do not attempt to install PySide, which must be part of FreeCAD already
            if QtCore.QThread.currentThread().isInterruptionRequested():
                return False
            try:
                proc = self._run_pip(
                    [
                        "install",
                        "--target",
                        vendor_path,
                        pymod,
                    ]
                )
                fci.Console.PrintMessage(proc.stdout + "\n")
            except subprocess.CalledProcessError as e:
                fci.Console.PrintError(str(e) + "\n")
                self.failure.emit(
                    translate(
                        "AddonsInstaller",
                        "Installation of Python package {} failed",
                    ).format(pymod),
                    str(e),
                )
                return False
        return True

    def _install_optional(self, vendor_path: str):
        """Install the optional Python package dependencies. If any fail a message is printed to
        the console, but installation of the others continues."""
        for pymod in self.python_optional:
            if QtCore.QThread.currentThread().isInterruptionRequested():
                return
            try:
                proc = self._run_pip(
                    [
                        "install",
                        "--target",
                        vendor_path,
                        pymod,
                    ]
                )
                fci.Console.PrintMessage(proc.stdout + "\n")
            except subprocess.CalledProcessError as e:
                fci.Console.PrintError(
                    translate("AddonsInstaller", "Installation of optional package failed")
                    + ":\n"
                    + str(e)
                    + "\n"
                )

    def _run_pip(self, args):
        final_args = utils.create_pip_call(args)
        return self._subprocess_wrapper(final_args)

    @staticmethod
    def _subprocess_wrapper(args) -> subprocess.CompletedProcess:
        """Wrap subprocess call so test code can mock it."""
        return utils.run_interruptable_subprocess(args, timeout_secs=120)

    def _install_addons(self):
        for addon in self.addons:
            if QtCore.QThread.currentThread().isInterruptionRequested():
                return
            fci.Console.PrintMessage(
                translate("AddonsInstaller", "Installing required dependency {}").format(addon.name)
                + "\n"
            )
            if addon.macro is None:
                installer = AddonInstaller(addon)
            else:
                installer = MacroInstaller(addon)
            result = installer.run()  # Run in this thread, which should be off the GUI thread
            if not result:
                self.failure.emit(
                    translate("AddonsInstaller", "Installation of addon {} failed").format(
                        addon.name
                    ),
                    "",
                )
                return



# ***************************************************************************
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com>   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Solver z88 FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import unittest
from os import listdir
from os.path import join
from os.path import isfile

import FreeCAD

import femsolver.run
from . import support_utils as testtools
from .support_utils import fcc_print
from .support_utils import get_namefromdef


class TestSolverZ88(unittest.TestCase):
    fcc_print("import TestSolverZ88")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

        # more inits
        self.mesh_name = "Mesh"

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars

        fcc_print(
            "\n{0}\n{1} run FEM TestSolverFrameWork tests {2}\n{0}".format(
                100 * "*", 10 * "*", 55 * "*"
            )
        )

    # ********************************************************************************************
    def test_ccx_cantilever_ele_hexa20(self):
        from femexamples.ccx_cantilever_ele_hexa20 import setup

        setup(self.document, "z88")
        self.inputfile_writing_test(get_namefromdef("test_"))

    # ********************************************************************************************
    def test_ccx_cantilever_ele_tria6(self):
        # TODO does pass on my local machine, but not on ci
        return

        from femexamples.ccx_cantilever_ele_tria6 import setup

        setup(self.document, "z88")
        self.inputfile_writing_test(get_namefromdef("test_"))

    # ********************************************************************************************
    def test_ccx_cantilever_faceload(self):
        from femexamples.ccx_cantilever_faceload import setup

        setup(self.document, "z88", test_mode=True)
        self.inputfile_writing_test(get_namefromdef("test_"))

    # ********************************************************************************************
    def test_ccx_cantilever_nodeload(self):
        from femexamples.ccx_cantilever_nodeload import setup

        setup(self.document, "z88", test_mode=True)
        self.inputfile_writing_test(get_namefromdef("test_"))

    # ********************************************************************************************
    def inputfile_writing_test(self, base_name):
        self.document.recompute()

        # start
        fcc_print(f"\n------------- Start of FEM Z88 tests for {base_name} -------")

        # get analysis working directory and save FreeCAD file
        working_dir = testtools.get_fem_test_tmp_dir("solver_z88_" + base_name)
        save_fc_file = join(working_dir, base_name + ".FCStd")
        fcc_print(f"Save FreeCAD file to {save_fc_file} ...")
        self.document.saveAs(save_fc_file)

        # write input file
        machine = self.document.SolverZ88.Proxy.createMachine(
            self.document.SolverZ88, working_dir, True  # set testmode to True
        )
        machine.target = femsolver.run.PREPARE
        machine.start()
        machine.join()  # wait for the machine to finish

        # compare created input files with the given input files
        test_path = join(testtools.get_fem_test_home_dir(), "z88", base_name)
        test_files = [f for f in listdir(test_path) if isfile(join(test_path, f))]
        # fcc_print(sorted(test_files))
        not_files = ["__init__.py"]
        test_files = [f for f in test_files if f not in not_files]
        fcc_print(test_files)
        for test_file in test_files:
            inpfile_given = join(test_path, test_file)
            inpfile_totest = join(working_dir, test_file)
            fcc_print(f"Comparing {inpfile_given}  to  {inpfile_totest}")
            ret = testtools.compare_inp_files(inpfile_given, inpfile_totest)
            self.assertFalse(ret, f"Z88 write_inp_file for {base_name} test failed.\n{ret}")

        # end
        fcc_print(f"--------------- End of FEM Z88 tests for {base_name} ---------")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


import argparse
import datetime
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
import FreeCAD
from FreeCAD import Units
import shlex
from builtins import open as pyopen

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for a smoothieboard. This postprocessor, once placed
in the appropriate PathScripts folder, can be used directly from inside
FreeCAD, via the GUI importer or via python scripts with:

import smoothie_post
smoothie_post.export(object,"/path/to/file.ncc","")
"""

now = datetime.datetime.now()

parser = argparse.ArgumentParser(prog="linuxcnc", add_help=False)
parser.add_argument("--header", action="store_true", help="output headers (default)")
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument("--comments", action="store_true", help="output comment (default)")
parser.add_argument("--no-comments", action="store_true", help="suppress comment output")
parser.add_argument("--line-numbers", action="store_true", help="prefix with line numbers")
parser.add_argument(
    "--no-line-numbers",
    action="store_true",
    help="don't prefix with line numbers (default)",
)
parser.add_argument(
    "--show-editor",
    action="store_true",
    help="pop up editor before writing output (default)",
)
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument("--precision", default="4", help="number of digits of precision, default=4")
parser.add_argument(
    "--preamble",
    help='set commands to be issued before the first command, default="G17\\nG90\\n"',
)
parser.add_argument(
    "--postamble",
    help='set commands to be issued after the last command, default="M05\\nG17 G90\\nM2\\n"',
)
parser.add_argument("--IP_ADDR", help="IP Address for machine target machine")
parser.add_argument(
    "--verbose",
    action="store_true",
    help='verbose output for debugging, default="False"',
)
parser.add_argument(
    "--inches", action="store_true", help="Convert output for US imperial mode (G20)"
)

TOOLTIP_ARGS = parser.format_help()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
IP_ADDR = None
VERBOSE = False

SPINDLE_SPEED = 0.0

if FreeCAD.GuiUp:
    SHOW_EDITOR = True
else:
    SHOW_EDITOR = False
MODAL = False  # if true commands are suppressed if the same as previous line.
COMMAND_SPACE = " "
LINENR = 100  # line number starting value

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
UNIT_SPEED_FORMAT = "mm/min"
UNIT_FORMAT = "mm"

MACHINE_NAME = "SmoothieBoard"
CORNER_MIN = {"x": 0, "y": 0, "z": 0}
CORNER_MAX = {"x": 500, "y": 300, "z": 300}

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = """G17 G90
"""

# Postamble text will appear following the last operation.
POSTAMBLE = """M05
G17 G90
M2
"""


# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = """"""

# Number of digits after the decimal point
PRECISION = 5


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global IP_ADDR
    global VERBOSE
    global PRECISION
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT

    try:
        args = parser.parse_args(shlex.split(argstring))

        if args.no_header:
            OUTPUT_HEADER = False
        if args.header:
            OUTPUT_HEADER = True
        if args.no_comments:
            OUTPUT_COMMENTS = False
        if args.comments:
            OUTPUT_COMMENTS = True
        if args.no_line_numbers:
            OUTPUT_LINE_NUMBERS = False
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        if args.show_editor:
            SHOW_EDITOR = True
        print("Show editor = %d" % SHOW_EDITOR)
        PRECISION = args.precision
        if args.preamble is not None:
            PREAMBLE = args.preamble.replace("\\n", "\n")
        if args.postamble is not None:
            POSTAMBLE = args.postamble.replace("\\n", "\n")
        if args.inches:
            UNITS = "G20"
            UNIT_SPEED_FORMAT = "in/min"
            UNIT_FORMAT = "in"

        IP_ADDR = args.IP_ADDR
        VERBOSE = args.verbose

    except Exception:
        return False

    return True


def export(objectslist, filename, argstring):
    processArguments(argstring)
    global UNITS
    for obj in objectslist:
        if not hasattr(obj, "Path"):
            FreeCAD.Console.PrintError(
                "the object "
                + obj.Name
                + " is not a path. Please select only path and Compounds.\n"
            )
            return

    FreeCAD.Console.PrintMessage("postprocessing...\n")
    gcode = ""

    # Find the machine.
    # The user my have overridden post processor defaults in the GUI.  Make
    # sure we're using the current values in the Machine Def.
    myMachine = None
    for pathobj in objectslist:
        if hasattr(pathobj, "MachineName"):
            myMachine = pathobj.MachineName
        if hasattr(pathobj, "MachineUnits"):
            if pathobj.MachineUnits == "Metric":
                UNITS = "G21"
            else:
                UNITS = "G20"
    if myMachine is None:
        FreeCAD.Console.PrintWarning("No machine found in this selection\n")

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(begin preamble)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"
    gcode += linenumber() + UNITS + "\n"

    for obj in objectslist:

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(begin operation: " + obj.Label + ")\n"
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(finish operation: " + obj.Label + ")\n"
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

    # do the post_amble

    if OUTPUT_COMMENTS:
        gcode += "(begin postamble)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    if SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    if IP_ADDR is not None:
        sendToSmoothie(IP_ADDR, final, filename)
    else:

        if not filename == "-":
            gfile = pyopen(filename, "w")
            gfile.write(final)
            gfile.close()

    FreeCAD.Console.PrintMessage("done postprocessing.\n")
    return final


def sendToSmoothie(ip, GCODE, fname):
    import sys
    import socket
    import os

    fname = os.path.basename(fname)
    FreeCAD.Console.PrintMessage("sending to smoothie: {}\n".format(fname))

    f = GCODE.rstrip()
    filesize = len(f)
    # make connection to sftp server
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(4.0)
    s.connect((ip, 115))
    tn = s.makefile(mode="rw")

    # read startup prompt
    ln = tn.readline()
    if not ln.startswith("+"):
        FreeCAD.Console.PrintMessage("Failed to connect with sftp: {}\n".format(ln))
        sys.exit()

    if VERBOSE:
        print("RSP: " + ln.strip())

    # Issue initial store command
    tn.write("STOR OLD /sd/" + fname + "\n")
    tn.flush()

    ln = tn.readline()
    if not ln.startswith("+"):
        FreeCAD.Console.PrintError("Failed to create file: {}\n".format(ln))
        sys.exit()

    if VERBOSE:
        print("RSP: " + ln.strip())

    # send size of file
    tn.write("SIZE " + str(filesize) + "\n")
    tn.flush()

    ln = tn.readline()
    if not ln.startswith("+"):
        FreeCAD.Console.PrintError("Failed: {}\n".format(ln))
        sys.exit()

    if VERBOSE:
        print("RSP: " + ln.strip())

    cnt = 0
    # now send file
    for line in f.splitlines(1):
        tn.write(line)
        if VERBOSE:
            cnt += len(line)
            print("SND: " + line.strip())
            print(str(cnt) + "/" + str(filesize) + "\r", end="")

    tn.flush()

    ln = tn.readline()
    if not ln.startswith("+"):
        FreeCAD.Console.PrintError("Failed to save file: {}\n".format(ln))
        sys.exit()

    if VERBOSE:
        print("RSP: " + ln.strip())

    # exit
    tn.write("DONE\n")
    tn.flush()
    tn.close()

    FreeCAD.Console.PrintMessage("Upload complete\n")


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 10
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    global SPINDLE_SPEED

    out = ""
    lastcommand = None
    precision_string = "." + str(PRECISION) + "f"

    # params = ['X','Y','Z','A','B','I','J','K','F','S'] #This list control
    # the order of parameters
    # linuxcnc doesn't want K properties on XY plane  Arcs need work.
    params = ["X", "Y", "Z", "A", "B", "I", "J", "F", "S", "T", "Q", "R", "L"]

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(" + pathobj.Label + ")\n"

        for c in PathUtils.getPathWithPlacement(pathobj).Commands:
            outstring = []
            command = c.Name
            outstring.append(command)
            # if modal: only print the command if it is not the same as the
            # last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F":
                        if c.Name not in [
                            "G0",
                            "G00",
                        ]:  # linuxcnc doesn't use rapid speeds
                            speed = Units.Quantity(c.Parameters["F"], FreeCAD.Units.Velocity)
                            outstring.append(
                                param
                                + format(
                                    float(speed.getValueAs(UNIT_SPEED_FORMAT)),
                                    precision_string,
                                )
                            )
                    elif param == "T":
                        outstring.append(param + str(c.Parameters["T"]))
                    elif param == "S":
                        outstring.append(param + str(c.Parameters["S"]))
                        SPINDLE_SPEED = c.Parameters["S"]
                    else:
                        pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                        outstring.append(
                            param + format(float(pos.getValueAs(UNIT_FORMAT)), precision_string)
                        )
            if command in ["G1", "G01", "G2", "G02", "G3", "G03"]:
                outstring.append("S" + str(SPINDLE_SPEED))

            # store the latest command
            lastcommand = command

            # Check for Tool Change:
            if command == "M6":
                # if OUTPUT_COMMENTS:
                #     out += linenumber() + "(begin toolchange)\n"
                for line in TOOL_CHANGE.splitlines(True):
                    out += linenumber() + line

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                for w in outstring:
                    out += w + COMMAND_SPACE
                out = out.strip() + "\n"

        return out


# print(__name__ + " gcode postprocessor loaded.")



# SPDX-License-Identifier: LGPL-2.1-or-later

#***************************************************************************
#*   Copyright (c) 2021 Chris Hennes <chennes@pioneerlibrarysystem.org>    *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENSE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

import unittest
import FreeCAD
import OpenSCAD
import importCSG
import tempfile
import os
import math

from os.path import join

__title__ = "ImportCSG OpenSCAD App unit tests"
__author__ = "Chris Hennes"
__url__ = "https://www.freecad.org"


class TestImportCSG(unittest.TestCase):

    MODULE = 'test_importCSG' # file name without extension
    temp_dir = tempfile.TemporaryDirectory()


    def setUp(self):
        self.test_dir = join(FreeCAD.getHomePath(), "Mod", "OpenSCAD", "OpenSCADTest", "data")

    def test_open_scad(self):
        testfile = join(self.test_dir, "CSG.scad")
        doc = importCSG.open(testfile)

        # Doc should now contain three solids: a union, an intersection, and a difference
        union = doc.getObject("union")
        intersection = doc.getObject("intersection")
        difference = doc.getObject("difference")

        self.assertTrue (union is not None)
        self.assertTrue (intersection is not None)
        self.assertTrue (difference is not None)

        FreeCAD.closeDocument("CSG")

    def test_open_csg(self):
        testfile = join(self.test_dir, "CSG.csg")
        doc = importCSG.open(testfile)

        # Doc should now contain three solids: a union, an intersection, and a difference
        union = doc.getObject("union")
        intersection = doc.getObject("intersection")
        difference = doc.getObject("difference")

        self.assertTrue (union is not None)
        self.assertTrue (intersection is not None)
        self.assertTrue (difference is not None)

        FreeCAD.closeDocument("CSG")

    def utility_create_scad(self, scadCode, name):
        filename = self.temp_dir.name + os.path.sep + name + ".scad"
        print (f"Creating {filename}")
        f = open(filename,"w+")
        f.write(scadCode)
        f.close()
        return importCSG.open(filename)

    def test_import_sphere(self):
        doc = self.utility_create_scad("sphere(10.0);","sphere")
        sphere = doc.getObject("sphere")
        self.assertTrue (sphere is not None)
        self.assertTrue (sphere.Radius == 10.0)
        FreeCAD.closeDocument(doc.Name)

    def test_import_cylinder(self):
        doc = self.utility_create_scad("cylinder(50.0,d=10.0);","cylinder")
        cylinder = doc.getObject("cylinder")
        self.assertTrue (cylinder is not None)
        self.assertTrue (cylinder.Radius == 5.0)
        self.assertTrue (cylinder.Height == 50.0)
        FreeCAD.closeDocument(doc.Name)

    def test_import_cube(self):
        doc = self.utility_create_scad("cube([1.0,2.0,3.0]);","cube")
        cube = doc.getObject("cube")
        self.assertTrue (cube is not None)
        self.assertTrue (cube.Length == 1.0)
        self.assertTrue (cube.Width == 2.0)
        self.assertTrue (cube.Height == 3.0)
        FreeCAD.closeDocument(doc.Name)

    def test_import_circle(self):
        doc = self.utility_create_scad("circle(10.0);","circle")
        circle = doc.getObject("circle")
        self.assertTrue (circle is not None)
        self.assertTrue (circle.Radius == 10.0)
        FreeCAD.closeDocument(doc.Name)

    def test_import_square(self):
        doc = self.utility_create_scad("square([1.0,2.0]);","square")
        square = doc.getObject("square")
        self.assertTrue (square is not None)
        self.assertTrue (square.Length == 1.0)
        self.assertTrue (square.Width == 2.0)
        FreeCAD.closeDocument(doc.Name)

    def test_import_text(self):
        # This uses the DXF importer that may pop-up modal dialogs
        # if not all 3rd party libraries are installed
        if FreeCAD.GuiUp:
            return
        try:
            doc = self.utility_create_scad("text(\"X\");","text") # Keep it short to keep the test fast-ish
            text = doc.getObject("text")
            self.assertTrue (text is not None)
            FreeCAD.closeDocument(doc.Name)
        except Exception:
            return # We may not have the DXF importer available

        # Try a number with a set script:
        doc = self.utility_create_scad("text(\"2\",script=\"latin\");","two_text")
        text = doc.getObject("text")
        self.assertTrue (text is not None)
        FreeCAD.closeDocument(doc.Name)

        # Leave off the script (which is supposed to default to "latin")
        doc = self.utility_create_scad("text(\"1\");","one_text")
        text = doc.getObject("text")
        self.assertTrue (text is not None)
        FreeCAD.closeDocument(doc.Name)

    def test_import_polygon_nopath(self):
        doc = self.utility_create_scad("polygon(points=[[0,0],[100,0],[130,50],[30,50]]);","polygon_nopath")
        polygon = doc.getObject("polygon")
        self.assertTrue (polygon is not None)
        self.assertAlmostEqual (polygon.Shape.Area, 5000.0)
        FreeCAD.closeDocument(doc.Name)

    def test_import_polygon_path(self):
        doc = self.utility_create_scad("polygon([[0,0],[100,0],[130,50],[30,50]], paths=[[0,1,2,3]]);","polygon_path")
        wire = doc.ActiveObject # With paths, the polygon gets created as a wire...
        self.assertTrue (wire is not None)
        self.assertAlmostEqual (wire.Shape.Area, 5000.0)
        FreeCAD.closeDocument(doc.Name)

    def test_import_polyhedron(self):
        doc = self.utility_create_scad(
"""
polyhedron(
  points=[ [10,10,0],[10,-10,0],[-10,-10,0],[-10,10,0], // the four points at base
           [0,0,10]  ],                                 // the apex point
  faces=[ [0,1,4],[1,2,4],[2,3,4],[3,0,4],              // each triangle side
              [1,0,3],[2,1,3] ]                         // two triangles for square base
 );
""","polyhedron"
                )
        polyhedron = doc.ActiveObject # With paths, the polygon gets created as a wire...
        self.assertTrue (polyhedron is not None)
        self.assertAlmostEqual (polyhedron.Shape.Volume, 1333.3333, 4)
        FreeCAD.closeDocument(doc.Name)

    def test_import_difference(self):
        doc = self.utility_create_scad("difference() { cube(15, center=true); sphere(10); }", "difference")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 266.1323, 3)
        FreeCAD.closeDocument(doc.Name)

    def test_import_intersection(self):
        doc = self.utility_create_scad("intersection() { cube(15, center=true); sphere(10); }", "intersection")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 3108.8677, 3)
        FreeCAD.closeDocument(doc.Name)

    def test_import_union(self):
        doc = self.utility_create_scad("union() { cube(15, center=true); sphere(10); }", "union")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 4454.9224, 3)
        FreeCAD.closeDocument(doc.Name)

    def test_import_rotate_extrude(self):
        doc = self.utility_create_scad("rotate_extrude() translate([10, 0]) square(5);", "rotate_extrude_simple")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 1963.4954, 3)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("translate([0, 30, 0]) rotate_extrude() polygon( points=[[0,0],[8,4],[4,8],[4,12],[12,16],[0,20]] );", "rotate_extrude_no_hole")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 2412.7431, 3)
        FreeCAD.closeDocument(doc.Name)

        # Bug #4353 - https://tracker.freecad.org/view.php?id=4353
        doc = self.utility_create_scad("rotate_extrude($fn=4, angle=180) polygon([[0,0],[3,3],[0,3]]);", "rotate_extrude_low_fn")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 9.0, 5)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("rotate_extrude($fn=4, angle=-180) polygon([[0,0],[3,3],[0,3]]);", "rotate_extrude_low_fn_negative_angle")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 9.0, 5)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("rotate_extrude(angle=180) polygon([[0,0],[3,3],[0,3]]);", "rotate_extrude_angle")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 4.5*math.pi, 5)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("rotate_extrude(angle=-180) polygon([[0,0],[3,3],[0,3]]);", "rotate_extrude_negative_angle")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 4.5*math.pi, 5)
        FreeCAD.closeDocument(doc.Name)

    def test_import_linear_extrude(self):
        doc = self.utility_create_scad("linear_extrude(height = 20) square([20, 10], center = true);", "linear_extrude_simple")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 4000.000, 3)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("linear_extrude(height = 20, twist = 90) square([20, 10], center = true);", "linear_extrude_twist")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 4000.000, 2)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("linear_extrude(height = 20, scale = 0.2) square([20, 10], center = true);", "linear_extrude_scale")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        h = 20
        a1 = 20*10
        a2 = 20*0.2 * 10*0.2
        expected_volume = (h/3) * (a1+a2+math.sqrt(a1*a2))
        self.assertAlmostEqual (object.Shape.Volume, expected_volume, 3)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("linear_extrude(height = 20, twist = 180, scale=0.2) square([20, 10], center = true);", "linear_extrude_twist_scale")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, expected_volume, 2)
        FreeCAD.closeDocument(doc.Name)

    def test_import_rotate_extrude_file(self):
        # OpenSCAD doesn't seem to have this feature at this time (March 2021)
        pass

# There is a problem with the DXF code right now, it doesn't like this square.
#    def test_import_import_dxf(self):
#        testfile = join(self.test_dir, "Square.dxf").replace('\\','/')
#        doc = self.utility_create_scad("import(\"{}\");".format(testfile), "import_dxf");
#        object = doc.ActiveObject
#        self.assertTrue (object is not None)
#        FreeCAD.closeDocument(doc.Name)

    def test_import_import_stl(self):
        preferences = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD")
        transfer_mechanism = preferences.GetInt('transfermechanism',0)
        if transfer_mechanism == 2:
            print ("Cannot test STL import, communication with OpenSCAD is via pipes")
            print ("If either OpenSCAD or FreeCAD are installed as sandboxed packages,")
            print ("use of import is not possible.")
            return
        testfile = join(self.test_dir, "Cube.stl").replace('\\','/')
        doc = self.utility_create_scad("import(\"{}\");".format(testfile), "import_stl");
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        FreeCAD.closeDocument(doc.Name)

    def test_import_resize(self):
        doc = self.utility_create_scad("resize([2,2,2]) cube();", "resize_simple")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 8.000000, 6)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("resize([2,2,0]) cube();", "resize_with_zero")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 4.000000, 6)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("resize([2,0,0], auto=true) cube();", "resize_with_auto")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 8.000000, 6)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("resize([2,2,2]) cube([2,2,2]);", "resize_no_change")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 8.000000, 6)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad("resize([2,2,2]) cube([4,8,12]);", "resize_non_uniform")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Volume, 8.000000, 6)
        FreeCAD.closeDocument(doc.Name)

        # Make sure to test something that isn't just a box (where the bounding box is trivial)
        doc = self.utility_create_scad("""
resize(newsize = [0,0,10], auto = [0,0,0]) {
    sphere($fn = 96, $fa = 12, $fs = 2, r = 8.5);
}""", "resize_non_uniform_sphere")
        object = doc.ActiveObject
        self.assertTrue (object is not None)
        object.Shape.tessellate(0.025) # To ensure the bounding box calculation is correct
        self.assertAlmostEqual (object.Shape.BoundBox.XLength, 2*8.5, 1)
        self.assertAlmostEqual (object.Shape.BoundBox.YLength, 2*8.5, 1)
        self.assertAlmostEqual (object.Shape.BoundBox.ZLength, 10.0, 1)
        FreeCAD.closeDocument(doc.Name)

    def test_import_surface(self):
        preferences = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/OpenSCAD")
        transfer_mechanism = preferences.GetInt('transfermechanism',0)
        if transfer_mechanism == 2:
            print ("Cannot test .dat surface import, communication with OpenSCAD is via pipes")
            print ("If either OpenSCAD or FreeCAD are installed as sandboxed packages, use of")
            print ("import is not possible.")
            return
        # Workaround for absolute vs. relative path issue
        # Inside the OpenSCAD file an absolute path name to Surface.dat is used
        # but by using the OpenSCAD executable to create a CSG file it's converted
        # into a path name relative to the output filename.
        # In order to open the CAG file correctly the cwd must be temporarily changed
        with tempfile.TemporaryDirectory() as temp_dir:
            cwd = os.getcwd()
            os.chdir(temp_dir)
            try:
                testfile = join(self.test_dir, "Surface.dat").replace('\\','/')
                doc = self.utility_create_scad(f"surface(file = \"{testfile}\", center = true, convexity = 5);", "surface_simple_dat")
                object = doc.ActiveObject
                self.assertTrue (object is not None)
                self.assertAlmostEqual (object.Shape.Volume, 275.000000, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.XMin, -4.5, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.XMax, 4.5, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.YMin, -4.5, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.YMax, 4.5, 6)
                FreeCAD.closeDocument(doc.Name)

                testfile = join(self.test_dir, "Surface.dat").replace('\\','/')
                doc = self.utility_create_scad(f"surface(file = \"{testfile}\", convexity = 5);", "surface_uncentered_dat")
                object = doc.ActiveObject
                self.assertTrue (object is not None)
                self.assertAlmostEqual (object.Shape.Volume, 275.000000, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.XMin, 0, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.XMax, 9, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.YMin, 0, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.YMax, 9, 6)
                FreeCAD.closeDocument(doc.Name)

                testfile = join(self.test_dir, "Surface2.dat").replace('\\','/')
                doc = self.utility_create_scad(f"surface(file = \"{testfile}\", center = true, convexity = 5);", "surface_rectangular_dat")
                object = doc.ActiveObject
                self.assertTrue (object is not None)
                self.assertAlmostEqual (object.Shape.Volume, 24.5500000, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.XMin, -2, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.XMax, 2, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.YMin, -1.5, 6)
                self.assertAlmostEqual (object.Shape.BoundBox.YMax, 1.5, 6)
                FreeCAD.closeDocument(doc.Name)
            except:
                os.chdir(cwd)
                raise Exception
            else:
                os.chdir(cwd)

    def test_import_projection(self):
        base_shape = "linear_extrude(height=5,center=true,twist=90,scale=0.5){square([1,1],center=true);}"
        hole = "cube([0.25,0.25,6],center=true);"
        cut_shape = f"difference() {{ {base_shape} {hole} }}"

        doc = self.utility_create_scad(f"projection(cut=true) {base_shape}", "projection_slice_square")
        object = doc.getObject("projection_cut")
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Area, 0.75*0.75, 3)
        FreeCAD.closeDocument(doc.Name)

        doc = self.utility_create_scad(f"projection(cut=true) {cut_shape}", "projection_slice_square_with_hole")
        object = doc.getObject("projection_cut")
        self.assertTrue (object is not None)
        self.assertAlmostEqual (object.Shape.Area, 0.75*0.75 - 0.25*0.25, 3)
        FreeCAD.closeDocument(doc.Name)

        # Unimplemented functionality:

        # With cut=false, the twisted unit square projects to a circle of radius sqrt(0.5)
        #doc = self.utility_create_scad(f"projection(cut=false) {base_shape}", "projection_circle")
        #object = doc.getObject("projection")
        #self.assertTrue (object is not None)
        #self.assertAlmostEqual (object.Shape.Area, 2*math.pi*math.sqrt(2), 3)
        #FreeCAD.closeDocument(doc.Name)

        #doc = self.utility_create_scad(f"projection(cut=false) {cut_shape}", "projection_circle_with_hole")
        #object = doc.getObject("projection")
        #self.assertTrue (object is not None)
        #self.assertAlmostEqual (object.Shape.Area, 2*math.pi*math.sqrt(0.5) - 0.125, 3)
        #FreeCAD.closeDocument(doc.Name)

    def test_import_hull(self):
        pass

    def test_import_minkowski(self):
        pass

    def test_import_offset(self):
        pass

    def test_empty_union(self):
        content = """union() {
 color(c = [0.30, 0.50, 0.80, 0.50]) {
  union() {
   union() {
    union() {
     translate(v = [23.0, -9.50, 13.60]) {
      union() {
       difference() {
        difference() {
         difference() {
          difference() {
           difference() {
            difference() {
             difference() {
              union() {
               union() {
                union() {
                 union() {
                  union() {
                   union() {
                    union() {
                     union();
                     translate(v = [2.50, 2.50, 9.50]) {
                      cylinder(h = 19.0, r = 2.50, center = true, $fn = 100);
                     }
                    }
                    translate(v = [11.50, 2.50, 9.50]) {
                     cylinder(h = 19.0, r = 2.50, center = true, $fn = 100);
                    }
                   }
                   translate(v = [11.50, 6.30, 9.50]) {
                    cylinder(h = 19.0, r = 2.50, center = true, $fn = 100);
                   }
                  }
                  translate(v = [2.50, 6.30, 9.50]) {
                   cylinder(h = 19.0, r = 2.50, center = true, $fn = 100);
                  }
                 }
                 translate(v = [2.50, 0.0, 0.0]) {
                  cube(size = [9.0, 8.80, 19.0]);
                 }
                }
                translate(v = [0.0, 2.50, 0.0]) {
                 cube(size = [14.0, 3.80, 19.0]);
                }
               }
              }
              translate(v = [-1.0, 8.40, 3.50]) {
               cube(size = [30.0, 10.0, 12.0]);
              }
             }
             translate(v = [4.0, 4.0, -0.10]) {
              cylinder($fn = 30, h = 20.0, r = 1.750, r1 = 1.80);
             }
            }
            translate(v = [9.0, 4.40, -0.10]) {
             cylinder($fn = 30, h = 20.0, r = 2.150, r1 = 2.20);
            }
           }
           translate(v = [12.30, 2.10, -0.10]) {
            cube(size = [5.0, 2.20, 20.0]);
           }
          }
          translate(v = [1.90, 7.60, -0.10]) {
           cube(size = [2.20, 5.0, 20.0]);
          }
         }
         translate(v = [3.60, 7.20, 1.80]) {
          cube(size = [30.0, 10.0, 2.0]);
         }
        }
        translate(v = [0, 0, 13.40]) {
         translate(v = [3.60, 7.20, 1.80]) {
          cube(size = [30.0, 10.0, 2.0]);
         }
        }
       }
      }
     }
    }
   }
  }
 }
}"""
        doc = self.utility_create_scad(content, "empty_union")
        self.assertEqual (len(doc.RootObjects), 1)
        FreeCAD.closeDocument(doc.Name)

    def test_complex_fuse_no_placement(self):
        # Issue #7878 - https://github.com/FreeCAD/FreeCAD/issues/7878

        csg_data = """
group() {
    multmatrix([[1, 0, 0, 0], [0, 1, 0, -127], [0, 0, 1, -6], [0, 0, 0, 1]]) {
        union() {
            group() {
                difference() {
                    cube(size = [4, 106.538, 12], center = false);
                    group() {
                            polyhedron(points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1]], faces = [[0, 1, 2], [5, 4, 3], [3, 1, 0], [1, 3, 4], [0, 2, 3], [5, 3, 2], [4, 2, 1], [2, 4, 5]], convexity = 1);
                    }
                }
            }
            polyhedron(points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1]], faces = [[0, 1, 2], [5, 4, 3], [3, 1, 0], [1, 3, 4], [0, 2, 3], [5, 3, 2], [4, 2, 1], [2, 4, 5]], convexity = 1);
            polyhedron(points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1]], faces = [[0, 1, 2], [5, 4, 3], [3, 1, 0], [1, 3, 4], [0, 2, 3], [5, 3, 2], [4, 2, 1], [2, 4, 5]], convexity = 1);
        }
    }
    multmatrix([[1, 0, 0, 6.4], [0, 1, 0, -125], [0, 0, 1, -40], [0, 0, 0, 1]]) {
        difference() {
            cylinder($fn = 0, $fa = 12, $fs = 2, h = 80, r1 = 8, r2 = 8, center = false);
            multmatrix([[1, 0, 0, -14.4], [0, 1, 0, -8], [0, 0, 1, -5], [0, 0, 0, 1]]) {
                cube(size = [8, 16, 90], center = false);
            }
        }
    }
}
"""
        doc = self.utility_create_scad(csg_data, "complex-fuse")
        self.assertEqual (doc.RootObjects[0].Placement, FreeCAD.Placement())
        FreeCAD.closeDocument(doc.Name)



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver equation base object"
__author__ = "Markus Hovorka"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import FreeCAD

if FreeCAD.GuiUp:
    from pivy import coin


class BaseProxy:

    BaseType = "App::FeaturePython"

    def __init__(self, obj):
        obj.Proxy = self
        obj.addProperty("App::PropertyLinkSubList", "References", "Base", "", locked=True)

    def execute(self, obj):
        return True


class BaseViewProxy:

    def __init__(self, vobj):
        vobj.Proxy = self

    def attach(self, vobj):
        default = coin.SoGroup()
        vobj.addDisplayMode(default, "Default")

    def getDisplayModes(self, obj):
        "Return a list of display modes."
        modes = ["Default"]
        return modes

    def getDefaultDisplayMode(self):
        return "Default"

    def setDisplayMode(self, mode):
        return mode


class DeformationProxy(BaseProxy):
    pass


class DeformationViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationDeformation.svg"


class ElasticityProxy(BaseProxy):
    pass


class ElasticityViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationElasticity.svg"


class ElectricforceProxy(BaseProxy):
    pass


class ElectricforceViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationElectricforce.svg"


class ElectrostaticProxy(BaseProxy):
    pass


class ElectrostaticViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationElectrostatic.svg"


class FlowProxy(BaseProxy):
    pass


class FlowViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationFlow.svg"


class FluxProxy(BaseProxy):
    pass


class FluxViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationFlux.svg"


class HeatProxy(BaseProxy):
    pass


class HeatViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationHeat.svg"


class MagnetodynamicProxy(BaseProxy):
    pass


class MagnetodynamicViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationMagnetodynamic.svg"


class Magnetodynamic2DProxy(BaseProxy):
    pass


class Magnetodynamic2DViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationMagnetodynamic2D.svg"


class StaticCurrentProxy(BaseProxy):
    pass


class StaticCurrentViewProxy(BaseViewProxy):

    def getIcon(self):
        return ":/icons/FEM_EquationStaticCurrent.svg"


##  @}
